# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A definition of properties shared between `AsyncDatabase` &
`Database`.
"""

__all__ = ["DatabaseProperties"]


from aiootp._typing import Typing as t
from aiootp._constants import FILENAME_HASH_BYTES
from aiootp._exceptions import InvalidSHMAC, Ignore, TimestampExpired
from aiootp._paths import DatabasePath
from aiootp.generics import ByteIO
from aiootp.ciphers import Chunky2048, Slick256

from .dbdomains import DBDomains
from .dbkdf import DBKDF


class DatabaseProperties:
    """
    A definition of properties shared between `AsyncDatabase` & `Database`.
    """

    __slots__ = (
        "__dict__",
        "_cache",
        "_is_metatag",
        "_manifest",
        "_profile_tokens",
        "_root_kdf",
        "_root_salt",
        "_silent",
        "path",
    )

    IO = ByteIO
    InvalidSHMAC = InvalidSHMAC
    TimestampExpired = TimestampExpired

    _Cipher: type = Chunky2048
    _DBDomains: type = DBDomains
    _DBKDF: type = DBKDF
    _ProfileTokens: type
    _TokenCipher: type = Slick256

    _ROOT_SALT_BYTES: int = 24
    _ROOT_SALT_LEDGERNAME: str = "0"
    _METATAGS_LEDGERNAME: str = "1"

    _path: t.Path = DatabasePath()

    @classmethod
    def _encode_filename(cls, value: bytes) -> str:
        """
        Returns the received bytes-type `value` in base38 encoding.
        """
        return cls.IO.bytes_to_filename(value)

    @property
    def _root_filename(self) -> str:
        """
        Returns the filename string of database's manifest.
        """
        return self._encode_filename(
            self._root_kdf.shake_128(
                size=FILENAME_HASH_BYTES, aad=self._DBDomains.ROOT_FILENAME
            )
        )

    @property
    def _root_path(self) -> t.Path:
        """
        Returns a `pathlib.Path` object that points to the file that
        contains the manifest ledger.
        """
        return self.path / self._root_filename

    @property
    def _maintenance_records(self) -> t.Set[str]:
        """
        Returns the ledgernames of entries in the manifest that refer to
        administrative values used by objects to track and coordinate
        themselves internally.
        """
        return {self._ROOT_SALT_LEDGERNAME, self._METATAGS_LEDGERNAME}

    @property
    def tags(self) -> t.Set[str]:
        """
        Returns a list of all user-defined names for values stored in
        the database object.
        """
        manifest = self._manifest
        return {
            getattr(manifest, filename)
            for filename in self._maintenance_records.symmetric_difference(
                manifest
            )
        }

    @property
    def filenames(self) -> t.Set[str]:
        """
        Returns a list of all derived filenames of user-defined tags
        stored in the database object.
        """
        return self._maintenance_records.symmetric_difference(
            self._manifest.__dict__
        )

    @property
    def metatags(self) -> t.Set[str]:
        """
        Returns the list of metatags that a database contains.
        """
        return set(
            self._manifest.__dict__.get(self._METATAGS_LEDGERNAME, ())
        )

    def _filename(self, tag: str) -> str:
        """
        Derives the filename hash given a user-defined `tag`.
        """
        filename = self._root_kdf.shake_128(
            self._root_salt,
            tag.encode(),
            aad=self._DBDomains.FILENAME,
            size=FILENAME_HASH_BYTES,
        )
        return self._encode_filename(filename)

    def __contains__(self, tag: str) -> bool:
        """
        Checks the cache & manifest for the filename associated with the
        user-defined `tag`.
        """
        filename = self._filename(tag)
        return filename in self._manifest or filename in self._cache

    def __bool__(self) -> bool:
        """
        Returns `True` if the instance has any tags or metatags set.
        """
        return bool(
            (len(self) > 0) or self._manifest[self._METATAGS_LEDGERNAME]
        )

    def __len__(self) -> int:
        """
        Returns the number tags currently added to the database.
        """
        return len(self._manifest) - len(self._maintenance_records)

    def __setitem__(self, tag: str, data: t.JSONSerializable) -> None:
        """
        Allows users to add the value `data` under the name `tag`
        into the database.
        """
        filename = self._filename(tag)
        setattr(self._cache, filename, data)
        setattr(self._manifest, filename, tag)

    def __getitem__(self, tag: str) -> t.Union[bytes, t.JSONSerializable]:
        """
        Allows users to retrieve the value stored under the name `tag`
        from the database cache.
        """
        filename = self._filename(tag)
        if filename in self._cache:
            return getattr(self._cache, filename)

    def __delitem__(self, tag: str) -> None:
        """
        Allows users to delete the value stored under the name `tag`
        from the database.
        """
        filename = self._filename(tag)
        with Ignore(KeyError, AttributeError):
            del self._manifest[filename]
        with Ignore(KeyError, AttributeError):
            del self._cache[filename]
        with Ignore(FileNotFoundError):
            (self.path / filename).unlink()


module_api = dict(
    DatabaseProperties=t.add_type(DatabaseProperties),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Implements a synchronous transparently encrypted database.
"""

__all__ = ["Database"]


import json

from aiootp._typing import Typing as t
from aiootp._constants import DEFAULT_AAD, DEFAULT_TTL, MIN_KEY_BYTES
from aiootp._constants import BYTES_FLAG, BYTES_FLAG_SIZE
from aiootp._constants import FILENAME_HASH_BYTES, SHAKE_128_BLOCKSIZE
from aiootp._exceptions import DatabaseIssue, Issue, Ignore
from aiootp._paths import delete_salt_file
from aiootp.commons import Namespace
from aiootp.randoms import token_bytes
from aiootp.keygens.passcrypt.config import passcrypt_spec

from .profile_tokens import ProfileTokens
from .database_properties import DatabaseProperties


class Database(DatabaseProperties):
    """
    This class creates databases which enable the disk persistence of
    any bytes or JSON serializable native python data-types, with fully
    transparent encryption / decryption.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    key = aiootp.csprng()
    db = Database(key)

    # Elements in a database are organized by user-defined tags ->
    db["income"] = 32000

    # Databases can store any JSON serializable data ->
    db["dict"] = {0: 1, 2: 3, 4: 5}-------------------------
    db["lists"] = ["juice", ["nested juice"]]              |
                                                           |
    # As well as raw bytes ->                              |
    db["bytes"] = b"value..."                              |
                                                           |
    # Save changes to disk ->                              |
    db.save_database()                                     |
                                                           |
    # Clear data (& unsaved changes) from the cache ->     |
    db.clear_cache()                                       |
                                                           |
    # Retrieve items by their tags ->                      |
    db["dict"]                                             |
    >>> None  # oops, it's not in the cache!               |
                                                           |
    db.query_tag("dict", cache=True)                       V
    >>> {"0": 1, "2": 3, "4": 5}  # <----- JSON turns keys into strings

    # Create descendant databases using what are called metatags ->
    taxes = db.metatag("taxes")
    taxes[2020] = {"jan": 130.25, "feb": 163.23, "mar": 149.68}
    assert taxes == db.taxes
    assert taxes[2020] == db.taxes[2020]

    # Delete a descendant database ->
    db.delete_metatag("taxes")

    # Or,

    # Purge the filesystem of all parent & descendant files ->
    db.delete_database()
    """

    __slots__ = ()

    _ProfileTokens: type = ProfileTokens

    @classmethod
    def generate_profile(
        cls,
        # passcrypt credentials
        *credentials: bytes,
        username: bytes,
        passphrase: bytes,
        salt: bytes = b"",
        aad: bytes = DEFAULT_AAD,
        # passcrypt settings
        mb: int = passcrypt_spec.DEFAULT_MB,
        cpu: int = passcrypt_spec.DEFAULT_CPU,
        cores: int = passcrypt_spec.DEFAULT_CORES,
        # database keyword arguments
        path: t.OptionalPathStr = None,
        preload: bool = False,
    ) -> t.Cls:
        """
        Creates & loads a profile database for a user from the `tokens`
        passed in.

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        db = aiootp.Database.generate_profile(
            b"server_url",     # Any number of arguments can be passed
            b"email_address",  # here as additional, optional credentials.
            username=b"username",
            passphrase=b"passphrase",
            salt=b"optional salt keyword argument",
            mb=256,   # The passcrypt memory cost in Mebibytes (MiB)
            cpu=2,    # The computational complexity & number of iterations
            cores=8,  # How many parallel processes passcrypt will utilize
        )
        """
        tokens = ProfileTokens().generate(
            *credentials,
            username=username,
            passphrase=passphrase,
            salt=salt,
            aad=aad,
            mb=mb,
            cpu=cpu,
            cores=cores,
            path=path,
        )
        profile_db = cls(key=tokens.login_key, path=path, preload=preload)
        if not profile_db._root_path.is_file():
            profile_db.save_database()
        profile_db._profile_tokens = tokens.cleanup()
        return profile_db

    def __init__(
        self,
        key: bytes,
        *,
        preload: bool = False,
        path: t.OptionalPathStr = None,
        metatag: bool = False,
        silent: bool = True,
    ) -> None:
        """
        Sets a database object's basic cryptographic values derived from
        a `key` & opens up the associated administrative files. The
        `generate_profile` method would be a safer choice for opening a
        database if using a passphrase instead of a cryptographic key.

        `preload`: This boolean value tells the object to -- True --
            load all of the stored database values from the filesystem
            into the cache during initialization, or -- False -- skip
            the loading stage. This trades-off between RAM & latency.

        `path`: This value is the string or `Pathlib.Path` object that
            points to the filesystem location where the database files
            reside / will be saved. By default, stores values in the
            directory "db" relative to the package source code.

        `metatag`: This boolean value tells the class whether to
            prepare itself as a sub-database or not. This signals the
            possibility of distinct internal implementation differences.

        `silent`: This boolean value tells the class to surpress
            exceptions when loading files so that errors in the database
            don't prevent a user from logging in.
        """
        self._silent = silent
        self._cache = Namespace()
        self._manifest = Namespace()
        self.path = self._format_path(path)
        self._is_metatag = bool(metatag)
        self._initialize_keys(key)
        self._load_manifest()
        self._initialize_metatags()
        self.load_database(silent=silent, preload=preload)

    @classmethod
    def _format_path(cls, path: t.PathStr) -> t.Path:
        """
        Returns a `pathlib.Path` object to the user-specified `path`
        if given, else returns a copy of the default database directory
        `Path` object.
        """
        if path is None:
            return t.Path(cls._path).absolute()
        return t.Path(path).absolute()

    def _initialize_keys(self, key: bytes) -> None:
        """
        Derives the database's cryptographic root key material and the
        filename of the manifest ledger.
        """
        self._root_kdf = self._DBKDF(self._DBDomains.ROOT_KDF, key=key)

    def _open_manifest(self) -> t.JSONObject:
        """
        Loads an existing manifest file ledger from the filesystem.
        """
        ciphertext = self.IO.read(path=self._root_path)
        key = self._root_kdf.shake_128(
            aad=self._DBDomains.MANIFEST, size=SHAKE_128_BLOCKSIZE
        )
        return self._Cipher(key).json_decrypt(ciphertext)

    def _load_root_salt(self) -> bytes:
        """
        Returns the decoded raw bytes root salt from the manifest.
        """
        return self.IO.urlsafe_to_bytes(
            self._manifest[self._ROOT_SALT_LEDGERNAME]
        )

    def _generate_root_salt(self) -> bytes:
        """
        Returns a raw bytes random salt with length metadata appended to
        be used as the instance's root salt.
        """
        return token_bytes(self._ROOT_SALT_BYTES)

    def _install_root_salt(self, salt: bytes) -> None:
        """
        Stores in the manifest, with a URL-safe base64 encoding, the
        `salt` value as its root salt. This the source of entropy for
        the database which isn't derived from the user's login key.
        """
        self._manifest[self._ROOT_SALT_LEDGERNAME] = (
            self.IO.bytes_to_urlsafe(salt).decode()
        )

    def _load_manifest(self) -> None:
        """
        Initalizes the object with a new database file ledger or loads
        an existing one from the filesystem.
        """
        if self._root_path.exists():
            self._manifest = Namespace(self._open_manifest())
            self._root_salt = self._load_root_salt()
        else:
            self._manifest = Namespace()
            self._root_salt = self._generate_root_salt()
            self._install_root_salt(self._root_salt)

    def _initialize_metatags(self) -> None:
        """
        Initializes the values that organize database metatags, which
        are independent offspring of databases that are accessible by
        their parent.
        """
        if not self.metatags:
            self._manifest[self._METATAGS_LEDGERNAME] = []

    def load_tags(self, *, silent: bool = False) -> t.Self:
        """
        Specifically loads all of the database's tag values into the
        cache.
        """
        for tag in self.tags:
            self.query_tag(tag, silent=silent, cache=True)
        return self

    def load_metatags(
        self, *, preload: bool = True, silent: bool = False
    ) -> t.Self:
        """
        Specifically loads all of the database's metatag values into the
        cache. If the `preload` keyword argument is falsey then the
        metatag references are populated in the database's instance
        dictionary, but their internal values are not loaded.
        """
        for metatag in set(self.metatags):
            self.metatag(metatag, preload=preload, silent=silent)
        return self

    def load_database(
        self,
        *,
        silent: bool = False,
        manifest: bool = False,
        preload: bool = True,
    ) -> t.Self:
        """
        Does initial loading of the database. If `manifest` is `True`,
        then the instance's manifest is reloaded from disk & unsaved
        changes to the manifest are discarded. If `preload` is `True`,
        then the value of all the database's tags, & the value of all of
        its metatag's tags, are cached from the filesystem & unsaved
        changes are discarded. This enables up-to-date bracket lookup of
        tag values without needing to call the `query_tag` method,
        but could be quite costly in terms of memory if databases &
        their metatags contain large amounts of data.
        """
        if manifest:
            self._load_manifest()
        if preload:
            self.load_tags(silent=silent)
        self.load_metatags(preload=preload, silent=silent)
        return self

    def filename(self, tag: str) -> str:
        """
        Derives the filename hash given a user-defined `tag`.
        """
        filename = self._root_kdf.shake_128(
            self._root_salt,
            tag.encode(),
            aad=self._DBDomains.FILENAME,
            size=FILENAME_HASH_BYTES,
        )
        return self._encode_filename(filename)

    def bytes_encrypt(
        self,
        plaintext: bytes,
        *,
        filename: str = "",
        aad: bytes = DEFAULT_AAD,
    ) -> bytes:
        """
        Encrypts the `plaintext` bytes with keys specific to the
        `filename` value & returns the ciphertext bytes.
        """
        key = self._root_kdf.shake_128(
            self._DBDomains.CIPHER,
            self._root_salt,
            filename.encode(),
            aad=aad,
            size=MIN_KEY_BYTES,
        )
        return self._Cipher(key).bytes_encrypt(plaintext)

    def json_encrypt(
        self,
        plaintext: t.JSONSerializable,
        *,
        filename: str = "",
        aad: bytes = DEFAULT_AAD,
    ) -> bytes:
        """
        Encrypts the JSON serializable `plaintext` object with keys
        specific to the `filename` value & returns the ciphertext
        bytes.
        """
        key = self._root_kdf.shake_128(
            self._DBDomains.CIPHER,
            self._root_salt,
            filename.encode(),
            aad=aad,
            size=MIN_KEY_BYTES,
        )
        return self._Cipher(key).json_encrypt(plaintext)

    def make_token(
        self,
        plaintext: bytes,
        *,
        filename: str = "",
        aad: bytes = DEFAULT_AAD,
    ) -> bytes:
        """
        Encrypts the `plaintext` bytes with keys specific to the
        `filename` value & urlsafe base64 encodes the resulting
        ciphertext bytes.
        """
        key = self._root_kdf.shake_128(
            self._DBDomains.CIPHER,
            self._root_salt,
            filename.encode(),
            aad=aad,
            size=MIN_KEY_BYTES,
        )
        return self._TokenCipher(key).make_token(plaintext)

    def bytes_decrypt(
        self,
        ciphertext: bytes,
        *,
        filename: str = "",
        aad: bytes = DEFAULT_AAD,
        ttl: t.Optional[int] = DEFAULT_TTL,
    ) -> bytes:
        """
        Decrypts the `ciphertext` bytes with keys specific to the
        `filename` value & returns the plaintext bytes. `ttl` is the
        amount of seconds that dictate the allowable age of the
        decrypted message.
        """
        key = self._root_kdf.shake_128(
            self._DBDomains.CIPHER,
            self._root_salt,
            filename.encode(),
            aad=aad,
            size=MIN_KEY_BYTES,
        )
        return self._Cipher(key).bytes_decrypt(ciphertext, ttl=ttl)

    def json_decrypt(
        self,
        ciphertext: bytes,
        *,
        filename: str = "",
        aad: bytes = DEFAULT_AAD,
        ttl: t.Optional[int] = DEFAULT_TTL,
    ) -> t.Union[bytes, t.JSONSerializable]:
        """
        Decrypts the `ciphertext` bytes with keys specific to the
        `filename` value & JSON loads the resulting plaintext bytes.
        `ttl` is the amount of seconds that dictate the allowable age
        of the decrypted message.
        """
        key = self._root_kdf.shake_128(
            self._DBDomains.CIPHER,
            self._root_salt,
            filename.encode(),
            aad=aad,
            size=MIN_KEY_BYTES,
        )
        return self._Cipher(key).json_decrypt(ciphertext, ttl=ttl)

    def read_token(
        self,
        token: t.Base64URLSafe,
        *,
        filename: str = "",
        aad: bytes = DEFAULT_AAD,
        ttl: t.Optional[int] = DEFAULT_TTL,
    ) -> bytes:
        """
        Decrypts the urlsafe base64 encoded `token` with keys specific
        to the `filename` value & returns the plaintext bytes. `ttl`
        is the amount of seconds that dictate the allowable age of the
        decrypted message.
        """
        key = self._root_kdf.shake_128(
            self._DBDomains.CIPHER,
            self._root_salt,
            filename.encode(),
            aad=aad,
            size=MIN_KEY_BYTES,
        )
        return self._TokenCipher(key).read_token(token, ttl=ttl)

    def _save_ciphertext(self, filename: str, ciphertext: bytes) -> None:
        """
        Saves the encrypted value `ciphertext` in the database file
        called `filename`.
        """
        path = self.path / filename
        self.IO.write(path=path, data=ciphertext)

    def set_tag(
        self, tag: str, data: t.JSONSerializable, *, cache: bool = True
    ) -> t.Self:
        """
        Allows users to add the value `data` under the name `tag`
        into the database.
        """
        filename = self.filename(tag)
        setattr(self._cache, filename, data)
        setattr(self._manifest, filename, tag)
        if not cache:
            self.save_tag(tag, drop_cache=True)
        return self

    def _query_ciphertext(
        self, filename: str, *, silent: bool = False
    ) -> None:
        """
        Retrieves the value stored in the database which has the given
        `filename`.
        """
        try:
            path = self.path / filename
            return self.IO.read(path=path)
        except FileNotFoundError as error:
            if not silent:
                raise DatabaseIssue.file_not_found(filename) from error

    def query_tag(
        self,
        tag: str,
        *,
        ttl: t.Optional[int] = DEFAULT_TTL,
        silent: bool = False,
        cache: bool = False,
    ) -> t.Union[bytes, t.JSONSerializable]:
        """
        Allows users to retrieve the value stored under the name `tag`
        from the database.
        """
        filename = self.filename(tag)
        if filename in self._cache:
            return getattr(self._cache, filename)
        ciphertext = self._query_ciphertext(filename, silent=silent)
        if not ciphertext:
            return

        result = self.bytes_decrypt(ciphertext, filename=filename, ttl=ttl)
        if result[:BYTES_FLAG_SIZE] == BYTES_FLAG:
            result = result[BYTES_FLAG_SIZE:]  # Remove bytes value flag
        else:
            result = json.loads(result)
        if cache:
            setattr(self._cache, filename, result)
        return result

    def _delete_file(self, filename: str, *, silent: bool = False) -> None:
        """
        Deletes a file in the database directory by `filename`.
        """
        try:
            (self.path / filename).unlink()
        except FileNotFoundError as error:
            if not silent:
                raise error from None

    def pop_tag(
        self, tag: str, *, silent: bool = False
    ) -> t.Union[bytes, t.JSONSerializable]:
        """
        Returns a value from the database by it's `tag` & deletes the
        associated file in the database directory.
        """

        def track_failure(relay: Ignore) -> bool:
            return failures.append(relay.error) or True

        failures = []
        filename = self.filename(tag)
        value = self.query_tag(tag, cache=False, silent=True)
        with Ignore(KeyError, AttributeError, if_except=track_failure):
            del self._manifest[filename]
        with Ignore(KeyError, AttributeError, if_except=track_failure):
            del self._cache[filename]
        with Ignore(FileNotFoundError, if_except=track_failure):
            self._delete_file(filename)
        if (not silent) and (len(failures) == 3):
            raise DatabaseIssue.tag_file_doesnt_exist(tag)
        return value

    def rollback_tag(self, tag: str, *, cache: bool = False) -> t.Self:
        """
        Clears the new `tag` data from the cache which undoes any
        recent changes. If the `tag` data was never saved to disk,
        then removing it from the cache will prevent it from being
        saved in the database.
        """
        filename = self.filename(tag)
        if not (self.path / filename).is_file():
            with Ignore(AttributeError):
                delattr(self._manifest, filename)
        with Ignore(AttributeError):
            delattr(self._cache, filename)
        with Ignore(LookupError):
            self.query_tag(tag, cache=True) if cache else 0
        return self

    def clear_cache(self, *, metatags: bool = True) -> t.Self:
        """
        Clears all recent changes in the cache. By default `metatags`
        is truthy, which clears a database's metatag caches.
        """
        self._cache.__dict__.clear()
        if metatags:
            for metatag in self.metatags:
                getattr(self, metatag).clear_cache(metatags=metatags)
        return self

    def _metatag_key(self, tag: str) -> bytes:
        """
        Derives the metatag's database key given a user-defined `tag`.
        """
        return self._root_kdf.sha3_512(
            self._root_salt, tag.encode(), aad=self._DBDomains.METATAG_KEY
        )

    def metatag(
        self, tag: str, *, preload: bool = False, silent: bool = False
    ) -> t.Cls:
        """
        Allows a user to create offspring of a database instance to
        organize data by a name `tag` & domain separate cryptographic
        material. These descendants are accessible by dotted lookup from
        the parent database. Descendants are also synchronized by their
        parents automatically.

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        # Create a parent database ->
        key = aiootp.csprng()
        parent = Database(key)

        # Name the descendant database ->
        tag = "sub_database"
        offspring = await parent.ametatag(tag)

        # It is now accessible from the parent by the tag ->
        assert offspring == parent.sub_database
        """
        if hasattr(self, tag):
            if (
                issubclass(getattr(self, tag).__class__, self.__class__)
                and tag not in self.__class__.__dict__
            ):
                return self.__dict__[tag]
            raise Issue.cant_reassign_attribute(tag)  # fixed v0.23.9
        self.__dict__[tag] = self.__class__(
            key=self._metatag_key(tag),
            preload=preload,
            path=self.path,
            metatag=True,
            silent=silent,
        )
        if tag not in self.metatags:
            getattr(self._manifest, self._METATAGS_LEDGERNAME).append(tag)
        return self.__dict__[tag]

    def delete_metatag(self, tag: str) -> t.Self:
        """
        Removes the descendant database named `tag`.
        """
        if tag not in self.metatags:
            raise DatabaseIssue.no_existing_metatag(tag)
        self.metatag(tag).delete_database()
        self.__dict__.pop(tag)
        self._manifest[self._METATAGS_LEDGERNAME].remove(tag)
        return self

    def _nullify(self) -> None:
        """
        Clears the database's memory caches & instance variables of all
        values so a deleted database no longer makes changes to the
        filesystem.
        """
        self._manifest.__dict__.clear()
        self._cache.__dict__.clear()
        self.__dict__.clear()
        for base in self.__class__.__mro__:
            for attr in getattr(base, "__slots__", ()):
                if hasattr(self, attr):
                    delattr(self, attr)

    def _delete_profile_tokens(self) -> None:
        """
        Deletes the salt file that what created if this instance was
        initialized with the `generate_profile` classmethod.
        """
        if getattr(self, "_profile_tokens", None):
            salt_path = self._profile_tokens._salt_path
            if salt_path.is_file():
                delete_salt_file(salt_path)

    def delete_database(self) -> None:
        """
        Completely clears all of the entries in database instance & its
        associated files.
        """
        for metatag in self.metatags:
            self.metatag(metatag, preload=False).delete_database()
        for filename in self._manifest.__dict__:
            self._delete_file(filename, silent=True)
        self._delete_file(self._root_filename, silent=True)
        self._delete_profile_tokens()
        self._nullify()

    def _encrypt_manifest(self) -> bytes:
        """
        Takes a `salt` & returns the database's manifest encrypted.
        """
        manifest = self._manifest.__dict__
        key = self._root_kdf.shake_128(
            aad=self._DBDomains.MANIFEST, size=SHAKE_128_BLOCKSIZE
        )
        return self._Cipher(key).json_encrypt(manifest)

    def _save_manifest(self, ciphertext: bytes) -> None:
        """
        Writes the manifest ledger to disk. It contains all database
        filenames & special cryptographic values for initializing the
        database's key derivation functions.
        """
        self.IO.write(path=self._root_path, data=ciphertext)

    def _close_manifest(self) -> None:
        """
        Prepares for & writes the manifest ledger to disk. The manifest
        contains all database filenames & other metadata used to
        organize databases.
        """
        manifest = self._encrypt_manifest()
        self._save_manifest(manifest)

    def _save_file(self, filename: str, *, admin: bool = False) -> None:
        """
        Writes the cached value for a user-specified `filename` to the
        user filesystem.
        """
        value = getattr(self._cache, filename)
        if value.__class__ is bytes:
            value = BYTES_FLAG + value  # Assure not reloaded as JSON
        else:
            value = json.dumps(value).encode()
        ciphertext = self.bytes_encrypt(value, filename=filename)
        self._save_ciphertext(filename, ciphertext)

    def _save_tags(self) -> None:
        """
        Writes the database's user-defined tags to disk.
        """
        save = self._save_file
        for filename in self._cache.__dict__:
            save(filename)

    def _save_metatags(self) -> None:
        """
        Writes the database's descendant databases to disk.
        """
        for metatag in self.metatags:
            getattr(self, metatag).save_database()

    def save_tag(
        self, tag: str, *, admin: bool = False, drop_cache: bool = False
    ) -> t.Self:
        """
        Writes the cached value for a user-specified `tag` to the user
        filesystem.
        """
        filename = self.filename(tag)
        try:
            self._save_file(filename, admin=admin)
        except AttributeError as error:
            raise DatabaseIssue.tag_file_doesnt_exist(tag) from error
        finally:
            if drop_cache and hasattr(self._cache, filename):
                delattr(self._cache, filename)
        return self

    def save_database(self) -> t.Self:
        """
        Writes the database's values to disk with transparent encryption.
        """
        self._close_manifest()
        self._save_metatags()
        self._save_tags()
        return self

    def __enter__(self) -> t.Self:
        """
        The context manager automatically writes database changes made
        by a user to disk.
        """
        return self

    def __exit__(
        self,
        /,
        exc_type: t.Optional[type] = None,
        exc_value: t.Optional[Exception] = None,
        traceback: t.Optional[t.TracebackType] = None,
    ) -> bool:
        """
        The context manager automatically writes database changes made
        by a user to disk.
        """
        self.save_database()
        return exc_type is None


module_api = dict(
    Database=t.add_type(Database),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
An interface for one of the package's online salt misuse-reuse
resistant, fully context commiting, tweakable, AEAD ciphers called
`Slick256`.
"""

__all__ = ["Slick256"]


from aiootp._typing import Typing as t

from .shake_permute_cipher import ShakePermuteKDFs
from .shake_permute_cipher import ShakePermuteKeyAADBundle
from .shake_permute_cipher import ShakePermuteStreamHMAC
from .shake_permute_cipher import ShakePermuteStreamJunction
from .cipher_interface import CipherInterface
from .slick_256_config import slick256_spec


class Slick256(CipherInterface):
    r"""
    An efficient high-level public interface to an online, salt misuse-
    reuse resistant, fully context committing, tweakable, RUP-secure,
    AEAD cipher.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    import aiootp

    key = aiootp.csprng()
    cipher = aiootp.Slick256(key)

    plaintext = b"Hello, Bob!"
    ciphertext = cipher.bytes_encrypt(plaintext)
    assert plaintext == cipher.bytes_decrypt(ciphertext)

    json_plaintext = ["any", {"JSON": "serializable object"}]
    ciphertext = cipher.json_encrypt(json_plaintext)
    assert json_plaintext == cipher.json_decrypt(ciphertext)

    token_plaintext = b"user_id|session_secret"
    token = cipher.make_token(token_plaintext)
    assert token_plaintext == cipher.read_token(token, ttl=3600)

     _____________________________________
    |                                     |
    |     Format Diagram: Ciphertext      |
    |_____________________________________|
     __________________________________________________________________
    |                       |                                          |
    |         Header        |                Ciphertext                |
    |---------|------|------|------|-------|-----------|---------|-----|
    |  shmac  | salt |  iv  | inner-header | plaintext |     footer    |
    |         |      |      |------|-------|           |---------|-----|
    |         |      |      | time | ikey  |           | padding | len |
    |    24   |   8  |   8  |   4  |   8   |     X     |    Y    |  1  |
    |_________|______|______|______|_______|___________|_________|_____|

     _____________________________________
    |                                     |
    |     Algorithm Pseudocode: Init      |
    |_____________________________________|

    e = a canonical, domain-specified encoding / padding function
    S = shmac_kdf = shake_128(e(SALT_S, METADATA, key, salt, aad, iv))
    K = primer_key = S.digest(336)
    π = permutation = FastAffineXORChain(key=K[168:168+π_key_size])
    P = 32-byte plaintext block
    C = 32-byte ciphertext block
    K_I, K_O, D = (K[:32], K[32:64], K[64:168])

    Each block is processed as such:

     _____________________________________
    |                                     |
    |    Algorithm Diagram: Encryption    |
    |_____________________________________|

                 K_I-------⊕--------       P
                /          ^       |       |                     /
               /           |       v       |                    /
         -----/            P     -----     v              -----/
    --->|  S  |                 |  π  |   (P ║ C ║ D)--->|  S  |
         -----\                  -----         ^          -----\
               \                   |           |                \
                \                  v           |                 \
                 K_O---------------⊕---------->C

     _____________________________________
    |                                     |
    |    Algorithm Diagram: Decryption    |
    |_____________________________________|

                 K_I---------------⊕------>P
                /                  ^       |                     /
               /                   |       |                    /
         -----/                  -----     v              -----/
    --->|  S  |                 |  π  |   (P ║ C ║ D)--->|  S  |
         -----\            C     -----         ^          -----\
               \           |       ^           |                \
                \          v       |           |                 \
                 K_O-------⊕--------           C
    """

    __slots__ = ()

    _KDFs: type = ShakePermuteKDFs
    _KeyAADBundle: type = ShakePermuteKeyAADBundle
    _StreamHMAC: type = ShakePermuteStreamHMAC
    _Junction: type = ShakePermuteStreamJunction

    _config: t.ConfigType = slick256_spec


module_api = dict(
    Slick256=t.add_type(Slick256),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A type for creating globally unique, pseudo-random identifiers
using bijective, keyed permutations.
"""

__all__ = ["GUID"]


from aiootp._typing import Typing as t
from aiootp._constants import BIG
from aiootp._exceptions import Issue
from aiootp._permutations import FastAffineXORChain
from aiootp.commons import ConfigMap, FrozenInstance

from .raw_guid import RawGUID, RawGUIDContainer
from .guid_config import GUIDConfig


class GUID(FrozenInstance):
    """
    A class for producing globally unique, pseudo-random identifiers
    that are guaranteed to be unique if all calls occur on a different
    nanosecond & use the same instance `key`. Additionally, any calls
    which utilize the same `key` but a different `node_id` will always
    produce unique outputs from each other, even if they occur on the
    same nanosecond. Outputs are also guaranteed to be unique if no more
    than 256 calls are made each nanosecond from any instance that
    utilizes the same `key` & `node_id`.

    Normal birthday-bound collision probabilities apply when a different
    `key` is used between callers.

    All outputs are masked using a pseudo-random permutation. They can
    be unmasked to reveal the raw GUIDs which by default are sorted by
    their `timestamp`, `node_id`, `ticker`, & `token`, in that order.

    By default 1-byte `node_id`s are supported, allowing 256 distinct
    guaranteed unique callers.

     _____________________________________
    |                                     |
    |   Format Diagram: Default Raw GUID  |
    |_____________________________________|
     __________________________________________________________________
    |                         |                    |         |         |
    |      timestamp (ns)     |       token        | node-id |  ticker |
    |-------------------------|--------------------|---------|---------|
    |         8-bytes         |       6-bytes      |  1-byte |  1-byte |
    |_________________________|____________________|_________|_________|
    |                                                                  |
    |                          16-bytes                                |
    |__________________________________________________________________|

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from io import StringIO
    from aiootp import GUID

    # Simple usage:

    def standard_guid_format(guid: bytes) -> str:
        read = StringIO(guid.hex()).read
        return "-".join(read(section) for section in (8, 4, 4, 4, 12))

    guids = GUID()
    print(standard_guid_format(guids.new()))
    8b36d9e4-cd73-8f67-631a-94e7b088c50a

    # Advanced usage:

    size = 16
    dds = distributed_database_system
    assert len(dds.shared_key) == GUID.key_size(size)
    guids = GUID(key=dds.shared_key, node_id=local_db.id)
    assert len(guids.new()) == size

    for table in local_db.tables:
        for record in table.new_records:
            record.dds_primary_key = guids.new()
            dds.merge(table, record)

    # Unmasking GUIDs

    guids = GUID(node_id=b"\xfc")
    print(guids.read(guids.new()))
    RawGUIDContainer(
        timestamp=b'\x00\xa3\xb9\xc3\xcc\x83CW',
        token=b'\xb4T<\xe7\xc8o',
        node_id=b'\xfc',
        ticker=b'\x01',
    )
    print(guids.read(guids.new()))
    RawGUIDContainer(
        timestamp=b'\x00\xa3\xb9\xc3\xcc\x88\x0f\x99',
        token=b'\xb5~\xf1\xf7\xa4\xab',
        node_id=b'\xfc',
        ticker=b'\x02',
    )
    """

    __slots__ = ("_raw_guid", "_permutation", "config")

    _configs = ConfigMap(
        mapping={
            config_id: GUIDConfig(
                config_id=config_id,
                size=config_id,
                raw_guid_type=RawGUID,
                permutation_type=FastAffineXORChain,
            )
            for config_id in range(12, 33)
        },
        config_type=GUIDConfig,
    )

    @classmethod
    def key_size(cls, config_id: t.Hashable) -> bytes:
        """
        Returns the number of bytes a uniform random key needs to be
        to initialize the instance's permutation.
        """
        return cls._configs[config_id].KEY_SIZE

    def _initialize_permutation(self, key: bytes) -> t.PermutationType:
        """
        Returns a bijective, keyed permutation as specified by the
        instance's configuration.
        """
        config_id = self.config.PERMUTATION_CONFIG_ID
        return self.config.Permutation(key=key, config_id=config_id)

    def __init__(
        self,
        *,
        key: t.Optional[bytes] = None,
        node_id: bytes = b"\x00",
        config_id: t.Hashable = 16,
    ) -> None:
        self.config = c = self._configs[config_id]
        key = key if key else c._DEFAULT_KEY
        self._raw_guid = c.RawGUID(node_id=node_id, config_id=config_id)
        self._permutation = self._initialize_permutation(key)

    async def anew(self) -> bytes:
        """
        Returns globally unique identifier that is blinded using a
        bijective, keyed permutation.
        """
        guid = await self._permutation.apermute(
            int.from_bytes(await self._raw_guid.anew(), BIG)
        )
        return guid.to_bytes(self.config.SIZE, BIG)

    def new(self) -> bytes:
        """
        Returns globally unique identifier that is blinded using a
        bijective, keyed permutation.
        """
        return self._permutation.permute(
            int.from_bytes(self._raw_guid.new(), BIG)
        ).to_bytes(self.config.SIZE, BIG)

    async def aread(self, guid: bytes) -> RawGUIDContainer:
        """
        Unblinds a `guid` & returns an object with the secret values
        parsed into it.
        """
        if len(guid) != self.config.SIZE:
            raise Issue.invalid_length("guid", len(guid))
        guid = int.from_bytes(guid, BIG)
        original = await self._permutation.ainvert(guid)
        return await self._raw_guid.aread(
            original.to_bytes(self.config.SIZE, BIG)
        )

    def read(self, guid: bytes) -> RawGUIDContainer:
        """
        Unblinds a `guid` & returns an object with the secret values
        parsed into it.
        """
        if len(guid) != self.config.SIZE:
            raise Issue.invalid_length("guid", len(guid))
        guid = int.from_bytes(guid, BIG)
        original = self._permutation.invert(guid)
        return self._raw_guid.read(original.to_bytes(self.config.SIZE, BIG))


module_api = dict(
    GUID=t.add_type(GUID),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
An interface for the package's online, salt misuse-reuse resistant,
"""

__all__ = ["CipherInterface"]


import json

from aiootp._typing import Typing as t
from aiootp._constants.misc import DEFAULT_AAD, DEFAULT_TTL
from aiootp._exceptions import Issue, TimestampExpired
from aiootp._exceptions import InvalidBlockID, InvalidSHMAC
from aiootp._gentools import abatch, batch
from aiootp.commons import FrozenInstance
from aiootp.generics import ByteIO

from .ciphertext_formatting import Ciphertext
from .padding import Padding
from .cipher_streams import AsyncCipherStream, CipherStream
from .decipher_streams import AsyncDecipherStream, DecipherStream


class CipherInterface(FrozenInstance):
    """
    A general definition for a high-level cipher interface.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    import aiootp

    key = aiootp.csprng()
    cipher = aiootp.CipherInterfaceSubclass(key)

    plaintext = b"Hello, Bob!"
    ciphertext = cipher.bytes_encrypt(plaintext)
    assert plaintext == cipher.bytes_decrypt(ciphertext)

    json_plaintext = ["any", {"JSON": "serializable object"}]
    ciphertext = cipher.json_encrypt(json_plaintext)
    assert json_plaintext == cipher.json_decrypt(ciphertext)

    token_plaintext = b"user_id|session_secret"
    token = cipher.make_token(token_plaintext)
    assert token_plaintext == cipher.read_token(token, ttl=3600)

     _____________________________________
    |                                     |
    |     Format Diagram: Ciphertext      |
    |_____________________________________|
     __________________________________________________________________
    |                       |                                          |
    |         Header        |                Ciphertext                |
    |---------|------|------|------|-------|-----------|---------|-----|
    |  shmac  | salt |  iv  | inner-header | plaintext | padding | len |
    |         |      |      |------|-------|           |         |     |
    |         |      |      | time | ikey  |           |         |     |
    |_________|______|______|______|_______|___________|_________|_____|

    """

    __slots__ = ("_kdfs",)

    InvalidBlockID: type = InvalidBlockID
    InvalidSHMAC: type = InvalidSHMAC
    TimestampExpired: type = TimestampExpired

    _KDFs: type
    _KeyAADBundle: type
    _StreamHMAC: type
    _Junction: type
    _AsyncCipherStream: type = AsyncCipherStream
    _CipherStream: type = CipherStream
    _AsyncDecipherStream: type = AsyncDecipherStream
    _DecipherStream: type = DecipherStream
    _Ciphertext: type = Ciphertext

    _config: t.ConfigType
    _padding: t.PaddingType

    def __init_subclass__(cls, /, *a: t.Any, **kw: t.Any) -> None:
        """
        Populates class specific configuration during class definition
        to reduce instance initialization costs.
        """
        cls._padding = Padding(config=cls._config)
        super().__init_subclass__(*a, **kw)

    def __init__(self, /, key: bytes) -> None:
        """
        Manages an encryption `key` within the state of a set of SHA3
        hashing objects defined by the cipher. This allows for efficient
        use of arbitrary sized keys, amortizing initialization time &
        memory costs.
        """
        self._kdfs = self._KDFs(key, config=self._config)

    async def abytes_encrypt(
        self,
        /,
        data: bytes,
        *,
        salt: t.Optional[bytes] = None,
        aad: bytes = DEFAULT_AAD,
    ) -> bytes:
        """
        Returns the ciphertext of any bytes type `data` containing the
        SHMAC authentication tag, the salt, & the IV.

        `salt`: A [pseudo]random salt that may be supplied by the user. By
                default it's sent in the clear attached to the ciphertext.
                Thus it may simplify implementing efficient features, such
                as search or routing, though care must still be taken when
                considering how leaking such metadata may be harmful.

                Keeping this value constant is strongly discouraged. Though,
                the cipher's salt misuse-reuse resistance is ruled by the
                combination of the automatically incorporated `timestamp`,
                `iv`, & `siv_key`. The risk calculation starts with setting
                r = len(iv + siv_key) / 3. Then, all else staying constant,
                once 256**r messages are encrypted within a second, each
                additional encrypted message within that same second begins
                to have more than a 256**(-r) chance of generating a repeat
                context.

                See: https://github.com/rmlibre/aiootp/issues/16

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.
        """
        key_bundle = await self._KeyAADBundle(
            self._kdfs, salt=salt, aad=aad
        ).async_mode()
        shmac = self._StreamHMAC(key_bundle)._for_encryption()
        data = abatch(
            await self._padding.apad_plaintext(data),
            size=self._config.BLOCKSIZE,
        )
        ciphering = self._Junction.abytes_encipher(data, shmac=shmac)
        ciphertext = (
            b"".join([block async for block in ciphering]),
            key_bundle.iv,
            key_bundle.salt,
            await shmac.afinalize(),
        )
        return b"".join(ciphertext[::-1])

    def bytes_encrypt(
        self,
        /,
        data: bytes,
        *,
        salt: t.Optional[bytes] = None,
        aad: bytes = DEFAULT_AAD,
    ) -> bytes:
        """
        Returns the ciphertext of any bytes type `data` containing the
        SHMAC authentication tag, the salt, & the IV.

        `salt`: A [pseudo]random salt that may be supplied by the user. By
                default it's sent in the clear attached to the ciphertext.
                Thus it may simplify implementing efficient features, such
                as search or routing, though care must still be taken when
                considering how leaking such metadata may be harmful.

                Keeping this value constant is strongly discouraged. Though,
                the cipher's salt misuse-reuse resistance is ruled by the
                combination of the automatically incorporated `timestamp`,
                `iv`, & `siv_key`. The risk calculation starts with setting
                r = len(iv + siv_key) / 3. Then, all else staying constant,
                once 256**r messages are encrypted within a second, each
                additional encrypted message within that same second begins
                to have more than a 256**(-r) chance of generating a repeat
                context.

                See: https://github.com/rmlibre/aiootp/issues/16

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.
        """
        key_bundle = self._KeyAADBundle(
            self._kdfs, salt=salt, aad=aad
        ).sync_mode()
        shmac = self._StreamHMAC(key_bundle)._for_encryption()
        data = batch(
            self._padding.pad_plaintext(data), size=self._config.BLOCKSIZE
        )
        ciphertext = (
            b"".join(self._Junction.bytes_encipher(data, shmac=shmac)),
            key_bundle.iv,
            key_bundle.salt,
            shmac.finalize(),
        )
        return b"".join(ciphertext[::-1])

    async def abytes_decrypt(
        self,
        /,
        data: bytes,
        *,
        aad: bytes = DEFAULT_AAD,
        ttl: t.Optional[int] = DEFAULT_TTL,
    ) -> bytes:
        """
        Returns the plaintext bytes from the bytes ciphertext `data`. The
        `data` bytes contain the SHMAC authentication tag, the salt, &
        the IV.

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.

        `ttl`: An amount of seconds that dictate the allowable age of
                the decrypted message.
        """
        data = self._Ciphertext(data, config=self._config)
        key_bundle = await self._KeyAADBundle(
            self._kdfs, salt=data.salt, aad=aad, iv=data.iv
        ).async_mode()
        shmac = self._StreamHMAC(key_bundle)._for_decryption()
        ciphertext = abatch(data.ciphertext, size=self._config.BLOCKSIZE)
        deciphering = self._Junction.abytes_decipher(
            ciphertext, shmac=shmac
        )
        plaintext = b"".join([block async for block in deciphering])
        await shmac.afinalize()
        await shmac.atest_shmac(data.shmac)
        return await self._padding.adepad_plaintext(plaintext, ttl=ttl)

    def bytes_decrypt(
        self,
        /,
        data: bytes,
        *,
        aad: bytes = DEFAULT_AAD,
        ttl: t.Optional[int] = DEFAULT_TTL,
    ) -> bytes:
        """
        Returns the plaintext bytes from the bytes ciphertext `data`. The
        `data` bytes contain the SHMAC authentication tag, the salt, &
        the IV.

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.

        `ttl`: An amount of seconds that dictate the allowable age of
                the decrypted message.
        """
        data = self._Ciphertext(data, config=self._config)
        key_bundle = self._KeyAADBundle(
            self._kdfs, salt=data.salt, aad=aad, iv=data.iv
        ).sync_mode()
        shmac = self._StreamHMAC(key_bundle)._for_decryption()
        ciphertext = batch(data.ciphertext, size=self._config.BLOCKSIZE)
        plaintext = b"".join(
            self._Junction.bytes_decipher(ciphertext, shmac=shmac)
        )
        shmac.finalize()
        shmac.test_shmac(data.shmac)
        return self._padding.depad_plaintext(plaintext, ttl=ttl)

    async def ajson_encrypt(
        self,
        /,
        data: t.JSONSerializable,
        *,
        salt: t.Optional[bytes] = None,
        aad: bytes = DEFAULT_AAD,
    ) -> bytes:
        """
        Returns the ciphertext of any JSON serializable `data` containing
        the SHMAC authentication tag, the salt, & the IV.

        `salt`: A [pseudo]random salt that may be supplied by the user. By
                default it's sent in the clear attached to the ciphertext.
                Thus it may simplify implementing efficient features, such
                as search or routing, though care must still be taken when
                considering how leaking such metadata may be harmful.

                Keeping this value constant is strongly discouraged. Though,
                the cipher's salt misuse-reuse resistance is ruled by the
                combination of the automatically incorporated `timestamp`,
                `iv`, & `siv_key`. The risk calculation starts with setting
                r = len(iv + siv_key) / 3. Then, all else staying constant,
                once 256**r messages are encrypted within a second, each
                additional encrypted message within that same second begins
                to have more than a 256**(-r) chance of generating a repeat
                context.

                See: https://github.com/rmlibre/aiootp/issues/16

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.
        """
        return await self.abytes_encrypt(
            json.dumps(data).encode(), salt=salt, aad=aad
        )

    def json_encrypt(
        self,
        /,
        data: t.JSONSerializable,
        *,
        salt: t.Optional[bytes] = None,
        aad: bytes = DEFAULT_AAD,
    ) -> bytes:
        """
        Returns the ciphertext of any JSON serializable `data` containing
        the SHMAC authentication tag, the salt, & the IV.

        `salt`: A [pseudo]random salt that may be supplied by the user. By
                default it's sent in the clear attached to the ciphertext.
                Thus it may simplify implementing efficient features, such
                as search or routing, though care must still be taken when
                considering how leaking such metadata may be harmful.

                Keeping this value constant is strongly discouraged. Though,
                the cipher's salt misuse-reuse resistance is ruled by the
                combination of the automatically incorporated `timestamp`,
                `iv`, & `siv_key`. The risk calculation starts with setting
                r = len(iv + siv_key) / 3. Then, all else staying constant,
                once 256**r messages are encrypted within a second, each
                additional encrypted message within that same second begins
                to have more than a 256**(-r) chance of generating a repeat
                context.

                See: https://github.com/rmlibre/aiootp/issues/16

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.
        """
        return self.bytes_encrypt(
            json.dumps(data).encode(), salt=salt, aad=aad
        )

    async def ajson_decrypt(
        self,
        /,
        data: bytes,
        *,
        aad: bytes = DEFAULT_AAD,
        ttl: t.Optional[int] = DEFAULT_TTL,
    ) -> t.JSONSerializable:
        """
        Decrypts the bytes ciphertext `data` & returns the loaded JSON
        plaintext. The `data` bytes contain the SHMAC authentication tag,
        the salt, & the IV.

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.

        `ttl`: An amount of seconds that dictate the allowable age of
                the decrypted message.
        """
        return json.loads(await self.abytes_decrypt(data, aad=aad, ttl=ttl))

    def json_decrypt(
        self,
        /,
        data: bytes,
        *,
        aad: bytes = DEFAULT_AAD,
        ttl: t.Optional[int] = DEFAULT_TTL,
    ) -> t.JSONSerializable:
        """
        Decrypts the bytes ciphertext `data` & returns the loaded JSON
        plaintext. The `data` bytes contain the SHMAC authentication tag,
        the salt, & the IV.

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.

        `ttl`: An amount of seconds that dictate the allowable age of
                the decrypted message.
        """
        return json.loads(self.bytes_decrypt(data, aad=aad, ttl=ttl))

    async def amake_token(
        self, /, data: bytes, *, aad: bytes = DEFAULT_AAD
    ) -> bytes:
        """
        Encrypts the bytes `data` & returns a urlsafe encoded ciphertext
        token. The `token` contains the SHMAC authentication tag, the
        salt, & the IV.

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.
        """
        if data.__class__ is not bytes:
            raise Issue.value_must_be_type("token data", bytes)
        ciphertext = await self.abytes_encrypt(data, aad=aad)
        return await ByteIO.abytes_to_urlsafe(ciphertext)

    def make_token(
        self, /, data: bytes, *, aad: bytes = DEFAULT_AAD
    ) -> bytes:
        """
        Encrypts the bytes `data` & returns a urlsafe encoded ciphertext
        token. The `token` contains the SHMAC authentication tag, the
        salt, & the IV.

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.
        """
        if data.__class__ is not bytes:
            raise Issue.value_must_be_type("token data", bytes)
        ciphertext = self.bytes_encrypt(data, aad=aad)
        return ByteIO.bytes_to_urlsafe(ciphertext)

    async def aread_token(
        self,
        /,
        token: t.Base64URLSafe,
        *,
        aad: bytes = DEFAULT_AAD,
        ttl: t.Optional[int] = DEFAULT_TTL,
    ) -> bytes:
        """
        Decodes a ciphertext `token` & returns the decrypted token
        data. The `token` bytes contain the SHMAC authentication tag,
        the salt, & the IV.

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.

        `ttl`: An amount of seconds that dictate the allowable age of
                the decrypted message.
        """
        if token.__class__ is not bytes:
            token = token.encode()
        ciphertext = await ByteIO.aurlsafe_to_bytes(token)
        return await self.abytes_decrypt(ciphertext, aad=aad, ttl=ttl)

    def read_token(
        self,
        /,
        token: t.Base64URLSafe,
        *,
        aad: bytes = DEFAULT_AAD,
        ttl: t.Optional[int] = DEFAULT_TTL,
    ) -> bytes:
        """
        Decodes a ciphertext `token` & returns the decrypted token
        data. The `token` bytes contain the SHMAC authentication tag,
        the salt, & the IV.

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.

        `ttl`: An amount of seconds that dictate the allowable age of
                the decrypted message.
        """
        if token.__class__ is not bytes:
            token = token.encode()
        ciphertext = ByteIO.urlsafe_to_bytes(token)
        return self.bytes_decrypt(ciphertext, aad=aad, ttl=ttl)

    async def astream_encrypt(
        self, /, *, salt: t.Optional[bytes] = None, aad: bytes = DEFAULT_AAD
    ) -> AsyncCipherStream:
        """
        Returns an object to manage encrypting a stream of plaintext.

        `salt`: A [pseudo]random salt that may be supplied by the user. By
                default it's sent in the clear attached to the ciphertext.
                Thus it may simplify implementing efficient features, such
                as search or routing, though care must still be taken when
                considering how leaking such metadata may be harmful.

                Keeping this value constant is strongly discouraged. Though,
                the cipher's salt misuse-reuse resistance is ruled by the
                combination of the automatically incorporated `timestamp`,
                `iv`, & `siv_key`. The risk calculation starts with setting
                r = len(iv + siv_key) / 3. Then, all else staying constant,
                once 256**r messages are encrypted within a second, each
                additional encrypted message within that same second begins
                to have more than a 256**(-r) chance of generating a repeat
                context.

                See: https://github.com/rmlibre/aiootp/issues/16

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.
        """
        return await self._AsyncCipherStream(self, salt=salt, aad=aad)

    def stream_encrypt(
        self, /, *, salt: t.Optional[bytes] = None, aad: bytes = DEFAULT_AAD
    ) -> CipherStream:
        """
        Returns an object to manage encrypting a stream of plaintext.

        `salt`: A [pseudo]random salt that may be supplied by the user. By
                default it's sent in the clear attached to the ciphertext.
                Thus it may simplify implementing efficient features, such
                as search or routing, though care must still be taken when
                considering how leaking such metadata may be harmful.

                Keeping this value constant is strongly discouraged. Though,
                the cipher's salt misuse-reuse resistance is ruled by the
                combination of the automatically incorporated `timestamp`,
                `iv`, & `siv_key`. The risk calculation starts with setting
                r = len(iv + siv_key) / 3. Then, all else staying constant,
                once 256**r messages are encrypted within a second, each
                additional encrypted message within that same second begins
                to have more than a 256**(-r) chance of generating a repeat
                context.

                See: https://github.com/rmlibre/aiootp/issues/16

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.
        """
        return self._CipherStream(self, salt=salt, aad=aad)

    async def astream_decrypt(
        self,
        /,
        *,
        salt: bytes,
        aad: bytes = DEFAULT_AAD,
        iv: bytes,
        ttl: t.Optional[int] = DEFAULT_TTL,
    ) -> AsyncDecipherStream:
        """
        Returns an object to manage decrypting a stream of ciphertext.

        `salt`: A [pseudo]random salt that may be supplied by the user. By
                default it's sent in the clear attached to the ciphertext.
                Thus it may simplify implementing efficient features, such
                as search or routing, though care must still be taken when
                considering how leaking such metadata may be harmful.

                Keeping this value constant is strongly discouraged. Though,
                the cipher's salt misuse-reuse resistance is ruled by the
                combination of the automatically incorporated `timestamp`,
                `iv`, & `siv_key`. The risk calculation starts with setting
                r = len(iv + siv_key) / 3. Then, all else staying constant,
                once 256**r messages are encrypted within a second, each
                additional encrypted message within that same second begins
                to have more than a 256**(-r) chance of generating a repeat
                context.

                See: https://github.com/rmlibre/aiootp/issues/16

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.

        `iv`: An ephemeral, uniform, random value that's generated by
                the encryption algorithm.

        `ttl`: An amount of seconds that dictate the allowable age of
                the decrypted message.
        """
        return await self._AsyncDecipherStream(
            self, salt=salt, aad=aad, iv=iv, ttl=ttl
        )

    def stream_decrypt(
        self,
        /,
        *,
        salt: bytes,
        aad: bytes = DEFAULT_AAD,
        iv: bytes,
        ttl: t.Optional[int] = DEFAULT_TTL,
    ) -> DecipherStream:
        """
        Returns an object to manage decrypting a stream of ciphertext.

        `salt`: A [pseudo]random salt that may be supplied by the user. By
                default it's sent in the clear attached to the ciphertext.
                Thus it may simplify implementing efficient features, such
                as search or routing, though care must still be taken when
                considering how leaking such metadata may be harmful.

                Keeping this value constant is strongly discouraged. Though,
                the cipher's salt misuse-reuse resistance is ruled by the
                combination of the automatically incorporated `timestamp`,
                `iv`, & `siv_key`. The risk calculation starts with setting
                r = len(iv + siv_key) / 3. Then, all else staying constant,
                once 256**r messages are encrypted within a second, each
                additional encrypted message within that same second begins
                to have more than a 256**(-r) chance of generating a repeat
                context.

                See: https://github.com/rmlibre/aiootp/issues/16

        `aad`: An arbitrary bytes value that a user decides to categorize
                keystreams. It's authenticated as associated data & safely
                differentiates keystreams as a tweak when it's unique for
                each permutation of `key`, `salt`, & `iv`.

        `iv`: An ephemeral, uniform, random value that's generated by
                the encryption algorithm.

        `ttl`: An amount of seconds that dictate the allowable age of
                the decrypted message.
        """
        return self._DecipherStream(
            self, salt=salt, aad=aad, iv=iv, ttl=ttl
        )


module_api = dict(
    CipherInterface=t.add_type(CipherInterface),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *

from aiootp._constants.misc import SECONDS, MILLISECONDS, MICROSECONDS


class TestRawGUIDConfig:
    async def test_declared_size_must_be_sum_of_components(self) -> None:
        problem = (  # fmt: skip
            "An inconsistent `size` declaration was allowed."
        )
        for bad_size in (15, 17):
            with Ignore(ValueError, if_else=violation(problem)):
                t.RawGUIDConfig(
                    timestamp_bytes=8,
                    prf_bytes=6,
                    node_id_bytes=1,
                    ticker_bytes=1,
                    size=bad_size,
                )

    async def test_custom_clock_can_be_supplied(self) -> None:
        for size, timestamp_bytes, units in (
            (16, 4, SECONDS),
            (18, 6, MILLISECONDS),
            (20, 8, MICROSECONDS),
        ):
            config = t.RawGUIDConfig(
                timestamp_bytes=timestamp_bytes,
                prf_bytes=10,
                node_id_bytes=1,
                ticker_bytes=1,
                size=size,
                clock=t.Clock(units),
            )
            assert units == config.clock._units


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Dependency inversion & documentation support for types relevant to
the `generics` subpackage.
"""

__all__ = [
    "Base64URLSafe",
    "HasherType",
    "JSONArray",
    "JSONDeserializable",
    "JSONObject",
    "JSONSerializable",
    "Number",
    "PositiveRealNumber",
    "RealNumber",
    "SupportsAppendPop",
    "SupportsAppendPopleft",
    "SupportsPopleft",
    "XOFType",
]


from .interface import Typing as t


Base64URLSafe = t.Base64URLSafe = t.NewType(
    "Base64URLSafe", t.Union[str, bytes]
)


Number = t.Number = t.NewType("Number", t.Union[int, float, complex])
RealNumber = t.RealNumber = t.NewType("RealNumber", t.Union[int, float])
PositiveRealNumber = t.PositiveRealNumber = t.NewType(
    "PositiveRealNumber", t.Union[int, float]
)


_JSONSerializableNonContainerTypes = t.NewType(
    "_JSONSerializableNonContainerTypes",
    t.Union[str, float, int, bool, None],
)
_JSONSerializableBaseTypes = t.NewType(
    "_JSONSerializableBaseTypes",
    t.Union[dict, list, _JSONSerializableNonContainerTypes],
)
JSONArray = t.JSONArray = t.NewType(
    "JSONArray", t.List[_JSONSerializableBaseTypes]
)
JSONObject = t.JSONObject = t.NewType(
    "JSONObject", t.Dict[str, _JSONSerializableBaseTypes]
)
JSONSerializable = t.JSONSerializable = t.NewType(
    "JSONSerializable",
    t.Union[JSONObject, JSONArray, _JSONSerializableNonContainerTypes],
)
JSONDeserializable = t.JSONDeserializable = t.NewType(
    "JSONDeserializable", t.Union[str, bytes, bytearray]
)


@t.runtime_checkable
class SupportsPopleft(t.Protocol):
    def popleft(self) -> t.Any:
        pass  # pragma: no cover


@t.runtime_checkable
class SupportsAppendPop(t.Protocol):
    def append(self, obj: t.Any, /) -> None:
        pass  # pragma: no cover

    def pop(self, index: int = -1, /) -> t.Any:
        pass  # pragma: no cover


@t.runtime_checkable
class SupportsAppendPopleft(t.Protocol):
    def append(self, obj: t.Any, /) -> None:
        pass  # pragma: no cover

    def popleft(self) -> t.Any:
        pass  # pragma: no cover


@t.runtime_checkable
class HasherType(t.Protocol):
    def copy(self, /) -> t.Self:
        pass  # pragma: no cover

    def update(self, data: bytes, /) -> None:
        pass  # pragma: no cover

    def digest(self, /) -> bytes:
        pass  # pragma: no cover

    def hexdigest(self, /) -> str:
        pass  # pragma: no cover


@t.runtime_checkable
class XOFType(HasherType, t.Protocol):
    def digest(self, size: int, /) -> bytes:
        pass  # pragma: no cover

    def hexdigest(self, size: int, /) -> str:
        pass  # pragma: no cover


module_api = dict(
    Base64URLSafe=Base64URLSafe,
    HasherType=t.add_type(HasherType),
    JSONArray=JSONArray,
    JSONDeserializable=JSONDeserializable,
    JSONObject=JSONObject,
    JSONSerializable=JSONSerializable,
    Number=Number,
    PositiveRealNumber=PositiveRealNumber,
    RealNumber=RealNumber,
    SupportsAppendPop=t.add_type(SupportsAppendPop),
    SupportsAppendPopleft=t.add_type(SupportsAppendPopleft),
    SupportsPopleft=t.add_type(SupportsPopleft),
    XOFType=t.add_type(XOFType),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A generic configuration type for the dual-output SHAKE cipher
mode of operation.
"""

__all__ = ["DualOutputShakeCipherConfig"]


from math import ceil
from hashlib import shake_128
from secrets import token_bytes

from aiootp._typing import Typing as t
from aiootp._constants import NamespaceMapping, BIG
from aiootp._exceptions import UndefinedRequiredAttributes
from aiootp.asynchs import Clock
from aiootp.commons import Config
from aiootp.generics import canonical_pack

from .cipher_kdfs import ShakeConfig


class DualOutputShakeCipherConfig(Config):
    """
    Configuration framework for dual-ouput SHAKE type ciphers.
    """

    __slots__ = (
        "BLOCKSIZE",
        "BLOCK_ID_BYTES",
        "BLOCK_ID_KDF",
        "BLOCK_ID_KDF_CONFIG",
        "BLOCK_ID_SLICE",
        "CIPHERTEXT_SLICE",
        "EMBEDDED_CAPACITY_BYTES",
        "EMBEDDED_CIPHERTEXT_SLICE",
        "EMBEDDED_LEFT_CAPACITY_SLICE",
        "EMBEDDED_RIGHT_CAPACITY_SLICE",
        "EPOCH_NS",
        "FIRST_CONTENT_BYTES",
        "FIRST_CONTENT_SLICE",
        "FIRST_DIGEST_SLICE",
        "FIRST_KEY_SLICE",
        "HALF_BLOCKSIZE",
        "HEADER_BYTES",
        "HEADER_SLICE",
        "INNER_BODY_SLICE",
        "INNER_HEADER_BYTES",
        "INNER_HEADER_SLICE",
        "IV_BYTES",
        "IV_SLICE",
        "KDF_CONFIGS",
        "LEFT_KDF",
        "LEFT_KDF_BLOCKSIZE",
        "LEFT_KDF_CONFIG",
        "LEFT_RATCHET_KEY_SLICE",
        "MAX_BLOCK_ID_BYTES",
        "MIN_BLOCK_ID_BYTES",
        "MIN_PADDING_BLOCKS",
        "MIN_STREAM_QUEUE",
        "NAME",
        "PACKED_METADATA",
        "PACKETSIZE",
        "PADDING_FRAME",
        "PERMUTATION_CONFIG_ID",
        "PERMUTATION_KEY_BYTES",
        "PERMUTATION_KEY_SLICE",
        "PERMUTATION_TEST_VECTOR",
        "PRIMER_KEY_BYTES",
        "Permutation",
        "RIGHT_KDF",
        "RIGHT_KDF_BLOCKSIZE",
        "RIGHT_KDF_CONFIG",
        "RIGHT_RATCHET_KEY_SLICE",
        "SALT_BYTES",
        "SALT_SLICE",
        "SENTINEL_BYTES",
        "SENTINEL_SLICE",
        "SHMAC_BLOCKSIZE",
        "SHMAC_BYTES",
        "SHMAC_DOUBLE_BLOCKSIZE",
        "SHMAC_KDF",
        "SHMAC_KDF_CONFIG",
        "SHMAC_RESULT_SLICE",
        "SHMAC_SLICE",
        "SIV_DIGEST_SLICE",
        "SIV_KEY_BYTES",
        "SIV_KEY_SLICE",
        "TIMESTAMP_BYTES",
        "TIMESTAMP_SLICE",
        "TIME_UNIT",
        "clock",
    )

    slots_types: t.Mapping[str, type] = dict(
        BLOCKSIZE=int,
        BLOCK_ID_BYTES=int,
        BLOCK_ID_KDF=str,
        BLOCK_ID_KDF_CONFIG=ShakeConfig,
        BLOCK_ID_SLICE=slice,
        CIPHERTEXT_SLICE=slice,
        CONFIG_ID=bytes,
        EMBEDDED_CAPACITY_BYTES=int,
        EMBEDDED_CIPHERTEXT_SLICE=slice,
        EMBEDDED_LEFT_CAPACITY_SLICE=slice,
        EMBEDDED_RIGHT_CAPACITY_SLICE=slice,
        EPOCH_NS=int,
        FIRST_CONTENT_BYTES=int,
        FIRST_CONTENT_SLICE=slice,
        FIRST_DIGEST_SLICE=slice,
        FIRST_KEY_SLICE=slice,
        HALF_BLOCKSIZE=int,
        HEADER_BYTES=int,
        HEADER_SLICE=slice,
        INNER_BODY_SLICE=slice,
        INNER_HEADER_BYTES=int,
        INNER_HEADER_SLICE=slice,
        IV_BYTES=int,
        IV_SLICE=slice,
        KDF_CONFIGS=NamespaceMapping,
        LEFT_KDF=str,
        LEFT_KDF_BLOCKSIZE=int,
        LEFT_KDF_CONFIG=ShakeConfig,
        LEFT_RATCHET_KEY_SLICE=slice,
        MAX_BLOCK_ID_BYTES=int,
        MIN_BLOCK_ID_BYTES=int,
        MIN_PADDING_BLOCKS=int,
        MIN_STREAM_QUEUE=int,
        NAME=str,
        PACKED_METADATA=bytes,
        PACKETSIZE=int,
        PADDING_FRAME=int,
        PERMUTATION_CONFIG_ID=t.Hashable,
        PERMUTATION_KEY_BYTES=int,
        PERMUTATION_KEY_SLICE=slice,
        PERMUTATION_TEST_VECTOR=bytes,
        PRIMER_KEY_BYTES=int,
        Permutation=t.PermutationType,
        RIGHT_KDF=str,
        RIGHT_KDF_BLOCKSIZE=int,
        RIGHT_KDF_CONFIG=ShakeConfig,
        RIGHT_RATCHET_KEY_SLICE=slice,
        SALT_BYTES=int,
        SALT_SLICE=slice,
        SENTINEL_BYTES=int,
        SENTINEL_SLICE=slice,
        SHMAC_BLOCKSIZE=int,
        SHMAC_BYTES=int,
        SHMAC_DOUBLE_BLOCKSIZE=int,
        SHMAC_KDF=str,
        SHMAC_KDF_CONFIG=ShakeConfig,
        SHMAC_RESULT_SLICE=slice,
        SHMAC_SLICE=slice,
        SIV_DIGEST_SLICE=slice,
        SIV_KEY_BYTES=int,
        SIV_KEY_SLICE=slice,
        TIMESTAMP_BYTES=int,
        TIMESTAMP_SLICE=slice,
        TIME_UNIT=str,
        clock=t.ClockType,
    )

    def __init__(
        self,
        *,
        name: str,
        config_id: bytes,
        blocksize: int,
        min_padding_blocks: int,
        epoch_ns: int,
        time_unit: str,
        timestamp_bytes: int,
        siv_key_bytes: int,
        shmac_bytes: int,
        block_id_bytes: int,
        max_block_id_bytes: int,
        min_block_id_bytes: int,
        salt_bytes: int,
        iv_bytes: int,
        block_id_kdf: str,
        block_id_kdf_config: ShakeConfig,
        shmac_kdf: str,
        shmac_kdf_config: ShakeConfig,
        left_kdf: str,
        left_kdf_config: ShakeConfig,
        right_kdf: str,
        right_kdf_config: ShakeConfig,
        permutation_type: t.PermutationType,
    ) -> None:
        self.NAME = name
        self.CONFIG_ID = config_id
        self.BLOCKSIZE = blocksize
        self.MIN_PADDING_BLOCKS = min_padding_blocks
        self.EPOCH_NS = epoch_ns
        self.TIME_UNIT = time_unit
        self.TIMESTAMP_BYTES = timestamp_bytes
        self.SIV_KEY_BYTES = siv_key_bytes
        self.SHMAC_BYTES = shmac_bytes
        self.BLOCK_ID_BYTES = block_id_bytes
        self.MAX_BLOCK_ID_BYTES = max_block_id_bytes
        self.MIN_BLOCK_ID_BYTES = min_block_id_bytes
        self.SALT_BYTES = salt_bytes
        self.IV_BYTES = iv_bytes
        self.BLOCK_ID_KDF = block_id_kdf
        self.BLOCK_ID_KDF_CONFIG = block_id_kdf_config
        self.SHMAC_KDF = shmac_kdf
        self.SHMAC_KDF_CONFIG = shmac_kdf_config
        self.LEFT_KDF = left_kdf
        self.LEFT_KDF_CONFIG = left_kdf_config
        self.RIGHT_KDF = right_kdf
        self.RIGHT_KDF_CONFIG = right_kdf_config
        self.Permutation = permutation_type
        self._perform_dynamic_configuration_from_constants()
        self._perform_correctness_checks()
        self._ensure_all_attributes_have_been_defined()

    def _alias_primitives_values(self) -> None:
        """
        Maps configuration values of cipher primitives.
        """
        # fmt: off
        self.KDF_CONFIGS = NamespaceMapping(**{
            self.SHMAC_KDF: self.SHMAC_KDF_CONFIG,
            self.LEFT_KDF: self.LEFT_KDF_CONFIG,
            self.RIGHT_KDF: self.RIGHT_KDF_CONFIG,
        })
        # fmt: on
        self.SHMAC_BLOCKSIZE = self.SHMAC_KDF_CONFIG.blocksize
        self.SHMAC_DOUBLE_BLOCKSIZE = self.SHMAC_KDF_CONFIG.double_blocksize
        self.SHMAC_RESULT_SLICE = slice(-self.SHMAC_BYTES, None, 1)
        self.LEFT_KDF_BLOCKSIZE = self.LEFT_KDF_CONFIG.blocksize
        self.LEFT_RATCHET_KEY_SLICE = self.LEFT_KDF_CONFIG.key_slice
        self.RIGHT_KDF_BLOCKSIZE = self.RIGHT_KDF_CONFIG.blocksize
        self.RIGHT_RATCHET_KEY_SLICE = self.RIGHT_KDF_CONFIG.key_slice

    def _compute_inner_header_measurements(self) -> None:
        """
        Use the provided configuration constants to determine the values
        of dependent inner header constants.
        """
        self.clock = Clock(self.TIME_UNIT, epoch=self.EPOCH_NS)
        self.INNER_HEADER_BYTES = self.TIMESTAMP_BYTES + self.SIV_KEY_BYTES
        self.INNER_HEADER_SLICE = slice(0, self.INNER_HEADER_BYTES, 1)
        self.INNER_BODY_SLICE = slice(self.INNER_HEADER_BYTES, None, 1)
        self.TIMESTAMP_SLICE = slice(0, self.TIMESTAMP_BYTES, 1)
        self.SIV_KEY_SLICE = slice(
            self.TIMESTAMP_BYTES, self.INNER_HEADER_BYTES, 1
        )

    def _compute_header_measurements(self) -> None:
        """
        Dynamic initialization of header constants from configuration.
        """
        self.HEADER_BYTES = (
            self.SHMAC_BYTES + self.SALT_BYTES + self.IV_BYTES
        )
        self.HALF_BLOCKSIZE = self.BLOCKSIZE // 2
        self.HEADER_SLICE = slice(0, self.HEADER_BYTES, 1)
        self.BLOCK_ID_SLICE = slice(0, self.BLOCK_ID_BYTES, 1)
        self.SHMAC_SLICE = slice(0, self.SHMAC_BYTES, 1)
        self.SALT_SLICE = slice(
            self.SHMAC_BYTES, self.SHMAC_BYTES + self.SALT_BYTES, 1
        )
        self.IV_SLICE = slice(
            self.SHMAC_BYTES + self.SALT_BYTES, self.HEADER_BYTES, 1
        )

    def _compute_permutation_integration_measurements(self) -> None:
        """
        Dynamic initialization of constants for the inner header masking
        permutation.
        """
        self.FIRST_DIGEST_SLICE = slice(0, self.SHMAC_BLOCKSIZE, 1)
        self.PERMUTATION_CONFIG_ID = self.INNER_HEADER_BYTES
        self.PERMUTATION_KEY_BYTES = self.Permutation.key_size(
            self.PERMUTATION_CONFIG_ID  # TODO @rmlibre: test less than shmac blocksize
        )
        self.PERMUTATION_KEY_SLICE = slice(
            -self.PERMUTATION_KEY_BYTES, None, 1
        )
        self.PRIMER_KEY_BYTES = self.SHMAC_BLOCKSIZE * (
            1 + ceil(self.PERMUTATION_KEY_BYTES / self.SHMAC_BLOCKSIZE)
        )

    def _compute_ciphertext_measurements(self) -> None:
        """
        Initialization of remaining ciphertext formatting constants.
        """
        self.PACKETSIZE = self.BLOCK_ID_BYTES + self.BLOCKSIZE
        self.CIPHERTEXT_SLICE = slice(self.HEADER_BYTES, None, 1)
        self.MIN_STREAM_QUEUE = self.MIN_PADDING_BLOCKS + 1
        self.PADDING_FRAME = self.BLOCKSIZE * self.MIN_STREAM_QUEUE
        self.SENTINEL_BYTES = ceil(self.PADDING_FRAME / 256)
        self.SENTINEL_SLICE = slice(-self.SENTINEL_BYTES, None, 1)

    def _compute_first_block_measurements(self) -> None:
        """
        Initialization of constants for synthetic IV algorithm.
        """
        self.EMBEDDED_CAPACITY_BYTES = (
            self.SHMAC_DOUBLE_BLOCKSIZE - self.BLOCKSIZE
        ) // 2
        self.EMBEDDED_LEFT_CAPACITY_SLICE = slice(
            0, self.EMBEDDED_CAPACITY_BYTES, 1
        )
        self.EMBEDDED_RIGHT_CAPACITY_SLICE = slice(
            -self.EMBEDDED_CAPACITY_BYTES, None, 1
        )
        self.EMBEDDED_CIPHERTEXT_SLICE = slice(
            self.EMBEDDED_CAPACITY_BYTES, -self.EMBEDDED_CAPACITY_BYTES, 1
        )
        self.FIRST_CONTENT_BYTES = self.BLOCKSIZE - self.INNER_HEADER_BYTES
        self.FIRST_CONTENT_SLICE = slice(self.INNER_HEADER_BYTES, None, 1)
        self.FIRST_KEY_SLICE = slice(
            self.EMBEDDED_CAPACITY_BYTES + self.INNER_HEADER_BYTES // 2,
            -self.EMBEDDED_CAPACITY_BYTES - self.INNER_HEADER_BYTES // 2,
            1,
        )
        self.SIV_DIGEST_SLICE = slice(2 * self.INNER_HEADER_BYTES, None, 1)
        # fmt: off
        self.PERMUTATION_TEST_VECTOR = self.Permutation(
            key=shake_128(self.CONFIG_ID).digest(self.PERMUTATION_KEY_BYTES),
            config_id=self.PERMUTATION_CONFIG_ID,
        ).permute(0).to_bytes(self.INNER_HEADER_BYTES, BIG)
        # fmt: on

    def _construct_metadata_constant(self) -> None:
        """
        Causes ciphertexts & plaintexts to be distinct & scrambled for
        distinct configurations.

        IMPORTANT FOR SECURITY.
        See:
        https://eprint.iacr.org/2016/292.pdf
        https://csrc.nist.gov/csrc/media/Events/2023/third-workshop-on-
            block-cipher-modes-of-operation/documents/accepted-papers/
            Flexible%20Authenticated%20Encryption.pdf

        DO NOT OVERRIDE TO PROVIDE ITER-OP.
        """
        self.PACKED_METADATA = canonical_pack(
            self.NAME.encode(),
            self.CONFIG_ID,
            self.EPOCH_NS.to_bytes(16, BIG),
            self.BLOCKSIZE.to_bytes(2, BIG),
            self.MIN_PADDING_BLOCKS.to_bytes(2, BIG),
            str(self.SHMAC_SLICE).encode(),
            str(self.BLOCK_ID_SLICE).encode(),
            str(self.SALT_SLICE).encode(),
            str(self.IV_SLICE).encode(),
            str(self.TIMESTAMP_SLICE).encode(),
            str(self.SIV_KEY_SLICE).encode(),
            str(self.SENTINEL_SLICE).encode(),
            self.PRIMER_KEY_BYTES.to_bytes(2, BIG),
            str(self.FIRST_DIGEST_SLICE).encode(),
            self.PERMUTATION_TEST_VECTOR,
            str(self.PERMUTATION_CONFIG_ID).encode(),
            str(self.PERMUTATION_KEY_SLICE).encode(),
            str(self.EMBEDDED_LEFT_CAPACITY_SLICE).encode(),
            str(self.EMBEDDED_RIGHT_CAPACITY_SLICE).encode(),
            str(self.EMBEDDED_CIPHERTEXT_SLICE).encode(),
            str(self.FIRST_CONTENT_SLICE).encode(),
            str(self.FIRST_KEY_SLICE).encode(),
            str(self.SIV_DIGEST_SLICE).encode(),
            str(self.SHMAC_RESULT_SLICE).encode(),
            self.BLOCK_ID_KDF_CONFIG.pad,
            self.BLOCK_ID_KDF_CONFIG.hasher().name.encode(),
            str(self.BLOCK_ID_KDF_CONFIG.key_slice).encode(),
            self.SHMAC_KDF_CONFIG.pad,
            self.SHMAC_KDF_CONFIG.hasher().name.encode(),
            str(self.SHMAC_KDF_CONFIG.key_slice).encode(),
            self.LEFT_KDF_CONFIG.pad,
            self.LEFT_KDF_CONFIG.hasher().name.encode(),
            str(self.LEFT_KDF_CONFIG.key_slice).encode(),
            self.RIGHT_KDF_CONFIG.pad,
            self.RIGHT_KDF_CONFIG.hasher().name.encode(),
            str(self.RIGHT_KDF_CONFIG.key_slice).encode(),
            int_bytes=1,
        )

    def _prepare_kdf_factories(self) -> None:
        """
        Each object part of the authentication & keystream generator
        defined for the cipher is initialized & updated with the
        serialized metadata of this configuration.
        """
        config_id = self.CONFIG_ID
        packed_metadata = self.PACKED_METADATA
        self.BLOCK_ID_KDF_CONFIG.prepare_factory(config_id, packed_metadata)
        for kdf in self.KDF_CONFIGS.values():
            kdf.prepare_factory(config_id, packed_metadata)

    def _perform_dynamic_configuration_from_constants(self) -> None:
        """
        Perform an ordered dynamic setup based on the provided values.
        """
        self._alias_primitives_values()
        self._compute_inner_header_measurements()
        self._compute_header_measurements()
        self._compute_permutation_integration_measurements()
        self._compute_ciphertext_measurements()
        self._compute_first_block_measurements()
        self._construct_metadata_constant()
        self._prepare_kdf_factories()

    def _ensure_inner_header_size_is_even(self) -> None:
        """
        `SyntheticIV` calculations require the len(inner_header) to be
        even, otherwise plaintext bytes could be leaked when salt reuse
        / misuse resistance fails.
        """
        if self.INNER_HEADER_BYTES % 2:
            raise ValueError(
                "INNER_HEADER_BYTES *cannot* be an odd number!"
            )

    def _ensure_blocksize_is_even(self) -> None:
        """
        The keystream consists of a left & right KDF which produce equal
        length keys equal to half the blocksize. An odd number blocksize
        could leak plaintext as the keys may not sum to the blocksize.
        It also introduces an unnecessary bias.
        """
        if self.BLOCKSIZE % 2:
            raise ValueError("BLOCKSIZE *cannot* be an odd number!")

    def _ensure_blocksize_is_positive(self) -> None:
        """
        A negative or zero blocksize cannot make sense.
        """
        if self.BLOCKSIZE <= 0:
            raise ValueError("BLOCKSIZE *cannot* be 0 or negative!")

    def _ensure_left_and_right_output_sizes_are_equal(self) -> None:
        """
        The output KDFs have to produce equal digest outputs.
        """
        if self.LEFT_KDF_BLOCKSIZE != self.RIGHT_KDF_BLOCKSIZE:
            raise ValueError("left & right blocksizes *must* be equal!")

    def _ensure_extracted_entropy_less_than_kdf_blocksize(self) -> None:
        """
        The blocksize of the output KDFs is intimately connected with the
        arguments of secuirty for the cipher. Particularly, exposing an
        amount of bytes larger than their internal blocksize, to XOR with
        plaintext, can only have negative impacts on security.
        """
        left_kdf = self.LEFT_KDF_BLOCKSIZE
        right_kdf = self.RIGHT_KDF_BLOCKSIZE
        entropy_window = left_kdf + right_kdf
        if self.BLOCKSIZE > entropy_window:
            raise ValueError(
                "BLOCKSIZE *must* sum to <= the left + right blocksizes!"
            )

    def _ensure_min_block_id_size_isnt_larger_than_max(self) -> None:
        """
        Sanity check.
        """
        if self.MIN_BLOCK_ID_BYTES > self.MAX_BLOCK_ID_BYTES:
            raise ValueError(
                "The MIN_BLOCK_ID_BYTES mustn't be larger than the "
                "MAX_BLOCK_ID_BYTES."
            )

    def _ensure_inner_header_leaves_adequate_space(self) -> None:
        """
        The inner-header must fit in the first block, therefore the
        blocksize of the cipher cannot be smaller than it. But, the
        available space for the first portion of plaintext must be at
        least 16-bytes. This ensures the possibility space of possible
        plaintexts for even single-block messages is sufficiently large.
        """
        remainder = (
            self.BLOCKSIZE - self.INNER_HEADER_BYTES - self.SENTINEL_BYTES
        ) % self.BLOCKSIZE
        if remainder < 16 and remainder >= 0:
            raise ValueError(
                "BLOCKSIZE - INNER_HEADER - SENTINEL_BYTES *must* leave "
                "at least 16-bytes for the plaintext!",
                remainder,
            )

    def _ensure_siv_measurements_are_correct(self) -> None:
        """
        Sanity check that the SIV algorithm component measurements line
        up as intended.
        """
        block = token_bytes(self.BLOCKSIZE)
        header = block[self.INNER_HEADER_SLICE]
        masked_header = token_bytes(self.INNER_HEADER_BYTES)
        key = token_bytes(self.SHMAC_DOUBLE_BLOCKSIZE)
        l_capacity = key[self.EMBEDDED_LEFT_CAPACITY_SLICE]
        r_capacity = key[self.EMBEDDED_RIGHT_CAPACITY_SLICE]
        ciphertext = masked_header + (
            int.from_bytes(key[self.FIRST_KEY_SLICE], BIG)
            ^ int.from_bytes(block[self.FIRST_CONTENT_SLICE], BIG)
        ).to_bytes(self.FIRST_CONTENT_BYTES, BIG)
        if len(ciphertext) != self.BLOCKSIZE:
            raise ValueError(
                "Derived SIV measurements were invalid!"
            )  # pragma: no cover
        elif (
            len(l_capacity + ciphertext + r_capacity)
            != self.SHMAC_DOUBLE_BLOCKSIZE
        ):
            raise ValueError(
                "Derived SIV measurements were invalid!"
            )  # pragma: no cover
        elif (
            len(
                header
                + masked_header
                + token_bytes(self.SHMAC_BLOCKSIZE)[self.SIV_DIGEST_SLICE]
            )
            != self.SHMAC_BLOCKSIZE
        ):
            raise ValueError(
                "Derived SIV measurements were invalid!"
            )  # pragma: no cover

    def _perform_correctness_checks(self) -> None:
        """
        Perform validity checks for the provided values & subsequent
        dynamic setup.
        """
        self._ensure_inner_header_size_is_even()
        self._ensure_blocksize_is_even()
        self._ensure_blocksize_is_positive()
        self._ensure_left_and_right_output_sizes_are_equal()
        self._ensure_extracted_entropy_less_than_kdf_blocksize()
        self._ensure_inner_header_leaves_adequate_space()
        self._ensure_min_block_id_size_isnt_larger_than_max()
        self._ensure_inner_header_leaves_adequate_space()
        self._ensure_siv_measurements_are_correct()

    def _ensure_all_attributes_have_been_defined(self) -> None:
        """
        Raise an error if a defined attribute failed to be initialized.
        """
        undefined_attributes = [
            name for name in self.__slots__ if not hasattr(self, name)
        ]
        if undefined_attributes:
            raise UndefinedRequiredAttributes(*undefined_attributes)


module_api = dict(
    DualOutputShakeCipherConfig=t.add_type(DualOutputShakeCipherConfig),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


import platform

from conftest import *


class TestMnemonics:
    async def test_default_is_random(self) -> None:
        assert mnemonic() != await amnemonic()

    async def test_returns_list_of_bytes_words(self) -> None:
        phrase = mnemonic()
        assert phrase.__class__ is list
        assert all((word in WORD_LIST) for word in phrase)
        assert all((word.__class__ is bytes) for word in phrase)

        aphrase = await amnemonic()
        assert aphrase.__class__ is list
        assert all((word in WORD_LIST) for word in aphrase)
        assert all((word.__class__ is bytes) for word in aphrase)

    async def test_size_dictates_word_count(self) -> None:
        for size in range(6, 12):
            size_word_phrase = mnemonic(size=size)
            asize_word_phrase = await amnemonic(size=size)
            assert size == len(size_word_phrase)
            assert size == len(asize_word_phrase)

    async def test_using_passphrase_is_deterministic(self) -> None:
        is_mac_os_issue = lambda _: (platform.system() == "Darwin")

        with Ignore(ConnectionRefusedError, if_except=is_mac_os_issue):
            phrase = mnemonic(passphrase, **passcrypt_settings)
            aphrase = await amnemonic(passphrase, **passcrypt_settings)
            assert phrase == aphrase

    async def test_async_parameters_dictate_functionality(self) -> None:
        problem = (  # fmt: skip
            "Can supply passcrypt settings to async mnemonic when not given "
            "passphrase."
        )
        with Ignore(ValueError, if_else=violation(problem)) as relay:
            await amnemonic(**passcrypt_settings)
        assert "parameters are not used" in relay.error.args[0]

    async def test_sync_parameters_dictate_functionality(self) -> None:
        problem = (  # fmt: skip
            "Can supply passcrypt settings to sync mnemonic when not given "
            "passphrase."
        )
        with Ignore(ValueError, if_else=violation(problem)) as relay:
            mnemonic(**passcrypt_settings)
        assert "parameters are not used" in relay.error.args[0]


class TestDomainKDF:
    aad: bytes = b"associated data..."
    domain: bytes = b"domain..."
    data: bytes = b"data..."
    inputs: t.Iterable[bytes] = (csprng(), csprng(), csprng())
    key: bytes = csprng()
    kdf: DomainKDF = DomainKDF(domain, data, key=key)

    async def test_empty_updates_arent_allowed(self) -> None:
        problem = (  # fmt: skip
            "Empty KDF updates were allowed."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            await self.kdf.aupdate()
        with Ignore(ValueError, if_else=violation(problem)):
            self.kdf.update()

    async def test_update_alters_state_distinctly(self) -> None:
        aupdated_kdf = DomainKDF(self.domain, key=self.key)
        updated_kdf = aupdated_kdf.copy()
        assert await aupdated_kdf.asha3_512(
            aad=self.aad
        ) == updated_kdf.sha3_512(aad=self.aad)

        initialized_kdf = DomainKDF(self.domain, self.data, key=self.key)

        await aupdated_kdf.aupdate(self.data)
        assert await aupdated_kdf.asha3_512(
            aad=self.aad
        ) != initialized_kdf.sha3_512(aad=self.aad)

        updated_kdf.update(self.data)
        assert updated_kdf.sha3_512(
            aad=self.aad
        ) != initialized_kdf.sha3_512(aad=self.aad)

    async def test_same_inputs_produce_same_outputs(self) -> None:
        assert self.kdf.sha3_256(*self.inputs) == await self.kdf.asha3_256(
            *self.inputs
        )
        assert self.kdf.sha3_512(*self.inputs) == await self.kdf.asha3_512(
            *self.inputs
        )
        assert self.kdf.shake_128(
            *self.inputs, size=32
        ) == await self.kdf.ashake_128(*self.inputs, size=32)
        assert self.kdf.shake_256(
            *self.inputs, size=32
        ) == await self.kdf.ashake_256(*self.inputs, size=32)

    async def test_same_aad_produces_same_outputs(self) -> None:
        assert self.kdf.sha3_256(aad=self.aad) == await self.kdf.asha3_256(
            aad=self.aad
        )
        assert self.kdf.sha3_512(aad=self.aad) == await self.kdf.asha3_512(
            aad=self.aad
        )
        assert self.kdf.shake_128(
            aad=self.aad, size=32
        ) == await self.kdf.ashake_128(aad=self.aad, size=32)
        assert self.kdf.shake_256(
            aad=self.aad, size=32
        ) == await self.kdf.ashake_256(aad=self.aad, size=32)

    async def test_different_inputs_produce_different_outputs(self) -> None:
        assert self.kdf.sha3_256() != await self.kdf.asha3_256(*self.inputs)
        assert self.kdf.sha3_512() != await self.kdf.asha3_512(*self.inputs)
        assert self.kdf.shake_128(size=32) != await self.kdf.ashake_128(
            *self.inputs, size=32
        )
        assert self.kdf.shake_256(size=32) != await self.kdf.ashake_256(
            *self.inputs, size=32
        )

    async def test_different_aad_produces_different_outputs(self) -> None:
        assert self.kdf.sha3_256() != await self.kdf.asha3_256(aad=self.aad)
        assert self.kdf.sha3_512() != await self.kdf.asha3_512(aad=self.aad)
        assert self.kdf.shake_128(size=32) != await self.kdf.ashake_128(
            aad=self.aad, size=32
        )
        assert self.kdf.shake_256(size=32) != await self.kdf.ashake_256(
            aad=self.aad, size=32
        )


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *


class TestDBKDF:
    async def test_salt_label_makes_state_unique(self, database) -> None:
        key = csprng()
        domain = b"testing"
        kdf = DomainKDF(domain, key=key)
        dbkdf = type(database._root_kdf)(domain, key=key)
        for sync_method in (
            "sha3_256",
            "sha3_512",
            "shake_128",
            "shake_256",
        ):
            if "shake" in sync_method:
                assert getattr(kdf, sync_method)(size=64) != getattr(
                    dbkdf, sync_method
                )(size=64)
            else:
                assert (
                    getattr(kdf, sync_method)()
                    != getattr(dbkdf, sync_method)()
                )


class TestDatabaseCacheSystem:
    async def cached_data_remains_unchanged(self, db, subdb) -> None:
        assert plaintext_bytes
        assert plaintext_bytes.__class__ is bytes
        assert subdb[tag] == plaintext_bytes
        assert test_data
        assert test_data.__class__ is dict
        assert db[tag] == test_data

    async def clearing_cache_results_in_null_values(
        self, db, subdb
    ) -> None:
        if issubclass(db.__class__, AsyncDatabase):
            await db.aclear_cache(metatags=True)
        else:
            db.clear_cache(metatags=True)
        assert plaintext_bytes
        assert plaintext_bytes.__class__ is bytes
        assert subdb[tag] is None
        assert test_data.__class__ is dict
        assert db[tag] is None

    async def uncached_loading_from_disk_doesnt_change_data(
        self, db, subdb
    ) -> None:
        if issubclass(db.__class__, AsyncDatabase):
            assert (
                await subdb.aquery_tag(tag, cache=False) == plaintext_bytes
            )
            assert await db.aquery_tag(tag, cache=False) == test_data
        else:
            assert subdb.query_tag(tag, cache=False) == plaintext_bytes
            assert db.query_tag(tag, cache=False) == test_data
        assert subdb[tag] is None
        assert db[tag] is None

    async def cached_loading_from_disk_doesnt_change_data(
        self, db, subdb
    ) -> None:
        if issubclass(db.__class__, AsyncDatabase):
            assert (
                await subdb.aquery_tag(tag, cache=True) == plaintext_bytes
            )
            assert await db.aquery_tag(tag, cache=True) == test_data
        else:
            assert subdb.query_tag(tag, cache=True) == plaintext_bytes
            assert db.query_tag(tag, cache=True) == test_data
        assert subdb[tag] == plaintext_bytes
        assert db[tag] == test_data

    async def clear_cache_clears_metatags_when_instructed(
        self, db, subdb
    ) -> None:
        if issubclass(db.__class__, AsyncDatabase):
            assert (
                await subdb.aquery_tag(tag, cache=True) == plaintext_bytes
            )
            assert subdb[tag] == plaintext_bytes
            await db.aclear_cache(metatags=False)
            assert subdb[tag] == plaintext_bytes
            await db.aclear_cache(metatags=True)
            assert subdb[tag] is None
            assert (
                await subdb.aquery_tag(tag, cache=False) == plaintext_bytes
            )
        else:
            assert subdb.query_tag(tag, cache=True) == plaintext_bytes
            assert subdb[tag] == plaintext_bytes
            db.clear_cache(metatags=False)
            assert subdb[tag] == plaintext_bytes
            db.clear_cache(metatags=True)
            assert subdb[tag] is None
            assert subdb.query_tag(tag, cache=False) == plaintext_bytes

    async def test_async_cache_system(self, async_database) -> None:
        subdb = await async_database.ametatag(metatag)
        subdb[tag] = plaintext_bytes
        async_database[tag] = test_data.copy()
        await async_database.asave_database()
        await self.cached_data_remains_unchanged(async_database, subdb)
        await self.clearing_cache_results_in_null_values(
            async_database, subdb
        )
        await self.uncached_loading_from_disk_doesnt_change_data(
            async_database, subdb
        )
        await self.cached_loading_from_disk_doesnt_change_data(
            async_database, subdb
        )
        await self.clear_cache_clears_metatags_when_instructed(
            async_database, subdb
        )

    async def test_sync_cache_system(self, database) -> None:
        subdb = database.metatag(metatag)
        subdb[tag] = plaintext_bytes
        database[tag] = test_data.copy()
        database.save_database()
        await self.cached_data_remains_unchanged(database, subdb)
        await self.clearing_cache_results_in_null_values(database, subdb)
        await self.uncached_loading_from_disk_doesnt_change_data(
            database, subdb
        )
        await self.cached_loading_from_disk_doesnt_change_data(
            database, subdb
        )
        await self.clear_cache_clears_metatags_when_instructed(
            database, subdb
        )


class TestDatabaseInitialization:
    async def test_async_key_size_limits(self) -> None:
        problem = (  # fmt: skip
            "a key that's too small was allowed."
        )
        token = token_bytes(64)
        for size in (0, 1, 2, 4, 8, 16, 31, MIN_KEY_BYTES - 1):
            key = token[:size]
            async with Ignore(
                ValueError,
                if_else=violation(f"{problem} at {len(key)} bytes"),
            ):
                await AsyncDatabase(key)

    def test_sync_key_size_limits(self) -> None:
        problem = (  # fmt: skip
            "a key that's too small was allowed."
        )
        token = token_bytes(64)
        for size in (0, 1, 2, 4, 8, 16, 31, MIN_KEY_BYTES - 1):
            key = token[:size]
            with Ignore(
                ValueError,
                if_else=violation(f"{problem} at {len(key)} bytes"),
            ):
                Database(key)


class TestDatabases:
    async def test_async_reload_manifest_drops_uncommitted_changes(
        self, async_database: AsyncDatabase
    ) -> None:
        manifest = async_database._manifest
        manifest["change"] = True
        assert "change" in manifest

        await async_database.aload_database(manifest=True)
        assert "change" in manifest
        assert "change" not in async_database._manifest

    async def test_sync_reload_manifest_drops_uncommitted_changes(
        self, database: Database
    ) -> None:
        manifest = database._manifest
        manifest["change"] = True
        assert "change" in manifest

        database.load_database(manifest=True)
        assert "change" in manifest
        assert "change" not in database._manifest

    async def test_async_non_existent_tag_query_doesnt_throw_on_silent(
        self, async_database: AsyncDatabase
    ) -> None:
        tag = "non_existent"
        assert tag not in async_database
        assert await async_database.aquery_tag(tag, silent=True) is None

        problem = (  # fmt: skip
            "Non-existent tag query doesn't throw when silent flag not set."
        )
        with Ignore(LookupError, if_else=violation(problem)):
            await async_database.aquery_tag(tag)

    async def test_sync_non_existent_tag_query_doesnt_throws_on_silent(
        self, database: Database
    ) -> None:
        tag = "non_existent"
        assert tag not in database
        assert database.query_tag(tag, silent=True) is None

        problem = (  # fmt: skip
            "Non-existent tag query doesn't throw when silent flag not set."
        )
        with Ignore(LookupError, if_else=violation(problem)):
            database.query_tag(tag)

    async def test_async_pop_doesnt_throw_on_silent(
        self, async_database: AsyncDatabase
    ) -> None:
        tag = "test_tag"
        data = b"test_data..."
        assert tag not in async_database
        assert await async_database.apop_tag(tag, silent=True) is None

        await async_database.aset_tag(tag, data)
        await async_database.asave_database()
        assert await async_database.apop_tag(tag) == data

        problem = (  # fmt: skip
            "Pop tag doesn't throw when silent flag not set."
        )
        with Ignore(LookupError, if_else=violation(problem)):
            await async_database.apop_tag(tag)

    async def test_sync_pop_doesnt_throw_on_silent(
        self, database: Database
    ) -> None:
        tag = "test_tag"
        data = b"test_data..."
        assert tag not in database
        assert database.pop_tag(tag, silent=True) is None

        database.set_tag(tag, data)
        database.save_database()
        assert database.pop_tag(tag) == data

        problem = (  # fmt: skip
            "Pop tag doesn't throw when silent flag not set."
        )
        with Ignore(LookupError, if_else=violation(problem)):
            database.pop_tag(tag)

    async def test_async_rollback_returns_previously_committed_state(
        self, async_database: AsyncDatabase
    ) -> None:
        tag = "test_tag"
        data = ["test", "data"]
        await async_database.apop_tag(tag, silent=True)
        async_database[tag] = list(data)
        assert async_database[tag] == data

        async_database[tag].append("mutated")
        assert async_database[tag] == [*data, "mutated"]

        await async_database.arollback_tag(tag, cache=True)
        assert async_database[tag] is None

        async_database[tag] = list(data)
        await async_database.asave_database()
        assert async_database[tag] == data

        async_database[tag].append("mutated")
        assert async_database[tag] == [*data, "mutated"]

        await async_database.arollback_tag(tag, cache=True)
        assert async_database[tag] == data

    async def test_sync_rollback_returns_previously_committed_state(
        self, database: Database
    ) -> None:
        tag = "test_tag"
        data = ["test", "data"]
        database.pop_tag(tag, silent=True)
        database[tag] = list(data)
        assert database[tag] == data

        database[tag].append("mutated")
        assert database[tag] == [*data, "mutated"]

        database.rollback_tag(tag, cache=True)
        assert database[tag] is None

        database[tag] = list(data)
        database.save_database()
        assert database[tag] == data

        database[tag].append("mutated")
        assert database[tag] == [*data, "mutated"]

        database.rollback_tag(tag, cache=True)
        assert database[tag] == data

    async def test_async_filenames_shows_all_set_tags(
        self, async_database: AsyncDatabase
    ) -> None:
        data = None
        async with async_database as db:
            tags = [f"tag_{i}" for i in range(32)]
            filenames = [await db.afilename(tag) for tag in tags]
            for tag in tags:
                db[tag] = data
            assert db.filenames.issuperset(filenames)

    async def test_sync_filenames_shows_all_set_tags(
        self, database: Database
    ) -> None:
        data = None
        with database as db:
            tags = [f"tag_{i}" for i in range(32)]
            filenames = [db.filename(tag) for tag in tags]
            for tag in tags:
                db[tag] = data
            assert db.filenames.issuperset(filenames)

    async def test_async_metatag_cant_point_to_non_database_type(
        self, async_database: AsyncDatabase
    ) -> None:
        class NonDatabaseType:
            pass

        problem = (  # fmt: skip
            "A non-database type metatag was assessed as a metatag."
        )
        metatag = "tested_attribute"
        async_database.tested_attribute = NonDatabaseType()
        with Ignore(PermissionError, if_else=violation(problem)):
            await async_database.ametatag(metatag)
        del async_database.tested_attribute

    async def test_sync_metatag_cant_point_to_non_database_type(
        self, database: Database
    ) -> None:
        class NonDatabaseType:
            pass

        problem = (  # fmt: skip
            "A non-database type metatag was assessed as a metatag."
        )
        metatag = "tested_attribute"
        database.tested_attribute = NonDatabaseType()
        with Ignore(PermissionError, if_else=violation(problem)):
            database.metatag(metatag)
        del database.tested_attribute

    async def test_async_delete_non_existent_metatag_throws_error(
        self, async_database: AsyncDatabase
    ) -> None:
        problem = (  # fmt: skip
            "A deletion of a non-existent metatag didn't throw an error."
        )
        metatag = "tested_metatag"
        with Ignore(LookupError, if_else=violation(problem)):
            await async_database.adelete_metatag(metatag)

    async def test_sync_delete_non_existent_metatag_throws_error(
        self, database: Database
    ) -> None:
        problem = (  # fmt: skip
            "A deletion of a non-existent metatag didn't throw an error."
        )
        metatag = "tested_metatag"
        with Ignore(LookupError, if_else=violation(problem)):
            database.delete_metatag(metatag)

    async def test_async_save_non_existent_tag_throws_error(
        self, async_database: AsyncDatabase
    ) -> None:
        problem = (  # fmt: skip
            "A save of a non-existent tag didn't throw an error."
        )
        tag = "tested_tag"
        await async_database.apop_tag(tag, silent=True)
        with Ignore(LookupError, if_else=violation(problem)):
            await async_database.asave_tag(tag)

    async def test_sync_save_non_existent_tag_throws_error(
        self, database: Database
    ) -> None:
        problem = (  # fmt: skip
            "A save of a non-existent tag didn't throw an error."
        )
        tag = "tested_tag"
        database.pop_tag(tag, silent=True)
        with Ignore(LookupError, if_else=violation(problem)):
            database.save_tag(tag)

    async def test_async_delitem_removes_tags(
        self, async_database: AsyncDatabase
    ) -> None:
        tag = "tested_tag"
        data = b"tested_data..."
        async_database[tag] = data
        await async_database.asave_database()
        assert async_database[tag] == data
        assert (
            async_database._path / await async_database.afilename(tag)
        ).is_file()

        del async_database[tag]
        assert async_database[tag] is None
        assert not (
            async_database._path / await async_database.afilename(tag)
        ).is_file()

    async def test_sync_delitem_removes_tags(
        self, database: AsyncDatabase
    ) -> None:
        tag = "tested_tag"
        data = b"tested_data..."
        database[tag] = data
        database.save_database()
        assert database[tag] == data
        assert (database._path / database.filename(tag)).is_file()

        del database[tag]
        assert database[tag] is None
        assert not (database._path / database.filename(tag)).is_file()


def test_Database_instance(database) -> None:
    db = Database(key=key, preload=True)

    # basic database functionalities work the same across reloads
    assert db._root_kdf.sha3_256() == database._root_kdf.sha3_256()
    assert db._root_kdf.sha3_256() != database._root_kdf.sha3_256(
        aad=database._root_salt
    )
    assert db._root_salt == database._root_salt
    assert db._root_filename == database._root_filename
    assert db.filename(atag) == database.filename(atag)
    assert db._metatag_key(atag) == database._metatag_key(atag)


async def test_AsyncDatabase_instance(database) -> None:
    db = await AsyncDatabase(key=key, preload=True)

    # basic async database functionalities work the same across reloads
    assert db._root_kdf.sha3_256() == database._root_kdf.sha3_256()
    assert db._root_kdf.sha3_256() != database._root_kdf.sha3_256(
        aad=database._root_salt
    )
    assert db._root_salt == database._root_salt
    assert db._root_filename == database._root_filename
    assert await db.afilename(tag) == database.filename(tag)
    assert await db._ametatag_key(tag) == database._metatag_key(tag)


def test_database_ciphers(database) -> None:
    # database ciphertexts are unique
    db = database
    cipher = Chunky2048(key)
    c = cipher._config
    filename = db.filename(tag)
    encrypted_data = db.json_encrypt(test_data, filename=filename, aad=b"")
    db[tag] = test_data
    db.save_database()
    encrypted_file = db._query_ciphertext(filename)
    assert encrypted_file != encrypted_data
    assert encrypted_file[c.SHMAC_SLICE] != encrypted_data[c.SHMAC_SLICE]
    assert encrypted_file[c.SALT_SLICE] != encrypted_data[c.SALT_SLICE]
    assert encrypted_file[c.IV_SLICE] != encrypted_data[c.IV_SLICE]
    assert (
        encrypted_file[c.CIPHERTEXT_SLICE]
        != encrypted_data[c.CIPHERTEXT_SLICE]
    )

    # database ciphers recover json data correctly
    assert test_data == db.json_decrypt(
        encrypted_file, filename=filename, aad=b""
    )
    assert test_data == db.json_decrypt(
        encrypted_data, filename=filename, aad=b""
    )
    assert type(test_data) is dict

    # database ciphers recover bytes data correctly
    encrypted_binary_data = db.bytes_encrypt(plaintext_bytes, aad=b"test")
    decrypted_binary_data = db.bytes_decrypt(
        encrypted_binary_data, aad=b"test", ttl=30
    )
    assert decrypted_binary_data == plaintext_bytes
    assert type(encrypted_binary_data) is bytes
    assert type(decrypted_binary_data) is bytes

    # database ciphers recover token data correctly
    encrypted_token_data = db.make_token(plaintext_bytes, aad=b"test")
    decrypted_token_data = db.read_token(
        encrypted_token_data, aad=b"test", ttl=3600
    )
    assert decrypted_token_data == plaintext_bytes
    assert type(encrypted_token_data) is bytes
    assert type(decrypted_token_data) is bytes
    assert set(Tables.URL_SAFE.encode()).union(
        encrypted_token_data + b"%"
    ) == set(Tables.URL_SAFE.encode()).union(b"%")


async def test_async_database_ciphers(async_database) -> None:
    # async database ciphertexts are unique
    db = async_database
    cipher = Chunky2048(key)
    c = cipher._config
    filename = await db.afilename(tag)
    encrypted_data = await db.ajson_encrypt(
        test_data, filename=filename, aad=b""
    )
    db[tag] = test_data
    await db.asave_database()
    encrypted_file = await db._aquery_ciphertext(filename)
    assert encrypted_file != encrypted_data
    assert encrypted_file[c.SHMAC_SLICE] != encrypted_data[c.SHMAC_SLICE]
    assert encrypted_file[c.SALT_SLICE] != encrypted_data[c.SALT_SLICE]
    assert encrypted_file[c.IV_SLICE] != encrypted_data[c.IV_SLICE]
    assert (
        encrypted_file[c.CIPHERTEXT_SLICE]
        != encrypted_data[c.CIPHERTEXT_SLICE]
    )

    # async database ciphers recover json data correctly
    assert test_data == await db.ajson_decrypt(
        encrypted_file, filename=filename, aad=b""
    )
    assert test_data == await db.ajson_decrypt(
        encrypted_data, filename=filename, aad=b""
    )

    # async database ciphers recover bytes data correctly
    encrypted_binary_data = await db.abytes_encrypt(
        plaintext_bytes, aad=b"test"
    )
    decrypted_binary_data = await db.abytes_decrypt(
        encrypted_binary_data, aad=b"test", ttl=30
    )
    assert decrypted_binary_data == plaintext_bytes
    assert type(encrypted_binary_data) is bytes

    # async database ciphers recover token data correctly
    encrypted_token_data = await db.amake_token(
        plaintext_bytes, aad=b"test"
    )
    decrypted_token_data = await db.aread_token(
        encrypted_token_data, aad=b"test", ttl=3600
    )
    assert decrypted_token_data == plaintext_bytes
    assert type(encrypted_token_data) is bytes
    assert set(Tables.URL_SAFE.encode()).union(
        encrypted_token_data + b"%"
    ) == set(Tables.URL_SAFE.encode()).union(b"%")


async def test_async_tags_metatags() -> None:
    # new, empty databases are falsey
    async_database = await AsyncDatabase(key * 2, preload=True)
    assert not async_database
    assert not async_database.tags
    assert not async_database.metatags

    # empty databases with metatag sub-databases are truthy
    achild = await async_database.ametatag(ametatag, preload=True)
    assert async_database
    assert not async_database.tags
    assert async_database.metatags
    assert not achild
    assert not achild.tags
    assert not achild.metatags

    # non-empty databases are truthy
    async_database[tag] = plaintext_bytes
    assert async_database
    assert async_database.tags
    assert async_database.metatags
    assert not achild
    assert not achild.tags
    assert not achild.metatags

    # async databases retrieve their stored data uncorrupted
    await async_database.aset_tag("bytes", plaintext_bytes, cache=False)
    assert await async_database.aquery_tag("bytes") == plaintext_bytes

    # metatag references stored as parent attributes are identical to
    # the references returned by their (a)metatag methods
    achild[atag] = atest_data
    assert achild is async_database.aclients
    assert achild[atag] is async_database.aclients[atag]

    # exiting an async database's async context manager saves their
    # childrens' files to disk
    afilename = await achild.afilename(atag)
    assert not (achild.path / afilename).exists()
    async with async_database:
        pass
    assert (achild.path / afilename).exists()

    # metatags of equivalent but not identical database instances
    # contain equivalent but not identical stored data
    adb = await AsyncDatabase(key * 2, preload=True)
    assert async_database.aclients[atag] == adb.aclients[atag]
    assert async_database.aclients[atag] is not adb.aclients[atag]

    # all databases create children from the keys returned by their
    # _ametatag_key methods
    ametatag_key = await async_database._ametatag_key(ametatag)
    achild = await AsyncDatabase(ametatag_key, metatag=True, preload=True)
    assert achild[atag]
    assert achild[atag].__class__ is dict
    assert achild[atag] == async_database.aclients[atag]
    assert achild[atag] is not async_database.aclients[atag]

    # metatags are removed from the manifest after calling adelete_metatag
    assert ametatag in async_database.metatags
    await async_database.adelete_metatag(ametatag)
    assert ametatag not in async_database.metatags

    await async_database.adelete_database()


def test_sync_tags_metatags() -> None:
    # new, empty databases are falsey
    database = Database(key * 2, preload=True)
    assert not database
    assert not database.tags
    assert not database.metatags

    # empty databases with metatag sub-databases are truthy
    child = database.metatag(metatag, preload=True)
    assert database
    assert not database.tags
    assert database.metatags
    assert not child
    assert not child.tags
    assert not child.metatags

    # non-empty databases are truthy
    database[tag] = plaintext_bytes
    assert database
    assert database.tags
    assert database.metatags
    assert not child
    assert not child.tags
    assert not child.metatags

    # databases retrieve their stored data uncorrupted
    database.set_tag("bytes", plaintext_bytes, cache=False)
    assert database.query_tag("bytes") == plaintext_bytes

    # metatag references stored as parent attributes are identical to
    # the references returned by their (a)metatag methods
    child[tag] = test_data
    assert child is database.clients
    assert child[tag] is database.clients[tag]

    # exiting database's context manager saves their childrens' files to
    # disk
    filename = child.filename(tag)
    assert not (child.path / filename).exists()
    with database:
        pass
    assert (child.path / filename).exists()

    # metatags of equivalent but not identical database instances
    # contain equivalent but not identical stored data
    db = Database(key * 2, preload=True)
    assert database.clients[tag] == db.clients[tag]
    assert database.clients[tag] is not db.clients[tag]

    # all databases create children from the keys returned by their
    # _metatag_key methods
    metatag_key = database._metatag_key(metatag)
    child = Database(metatag_key, metatag=True, preload=True)
    assert child[tag]
    assert child[tag].__class__ is dict
    assert child[tag] == database.clients[tag]
    assert child[tag] is not database.clients[tag]

    # metatags are removed from the manifest after calling delete_metatag
    assert metatag in database.metatags
    database.delete_metatag(metatag)
    assert metatag not in database.metatags

    database.delete_database()


async def test_async_user_profiles(async_database) -> None:
    adb = async_database
    user = await adb.agenerate_profile(**PROFILE_AND_SETTINGS)

    async with user:
        user[atag] = atest_data

    # equivalent async profiles contain their own copies of stored data
    user_copy = await adb.agenerate_profile(
        **PROFILE_AND_SETTINGS, preload=True
    )
    assert user[atag] == user_copy[atag]
    assert user.tags == user_copy.tags
    assert user[atag]
    assert user[atag] == atest_data
    assert user[atag].__class__ is dict
    assert user[atag] == user_copy[atag]
    assert user[atag] is not user_copy[atag]

    # async profiles are automatically saved to disk when initialized
    assert user_copy._root_path.is_file()
    assert user_copy._profile_tokens._salt_path.is_file()
    assert (user_copy.path / await user_copy.afilename(atag)).is_file()

    # async & sync profile contructors are equivalent
    sync_user = Database.generate_profile(
        **PROFILE_AND_SETTINGS, preload=True
    )
    assert sync_user[atag] == user[atag]
    assert sync_user.tags == user.tags
    assert sync_user[atag]
    assert sync_user[atag] == atest_data
    assert sync_user[atag].__class__ is dict
    assert sync_user[atag] is not user[atag]

    # deleting an async profile removes its files from the filesystem
    await user.adelete_database()
    assert not user_copy._root_path.is_file()
    assert not user_copy._profile_tokens._salt_path.is_file()
    assert not (user_copy.path / await user_copy.afilename(atag)).is_file()


async def test_user_profiles(database) -> None:
    db = database
    user = db.generate_profile(**PROFILE_AND_SETTINGS)

    with user:
        user[tag] = test_data

    # equivalent profiles contain their own copies of stored data
    user_copy = db.generate_profile(**PROFILE_AND_SETTINGS, preload=True)
    assert user[tag] == user_copy[tag]
    assert user.tags == user_copy.tags
    assert user[tag]
    assert user[tag] == atest_data
    assert user[tag].__class__ is dict
    assert user[tag] == user_copy[tag]
    assert user[tag] is not user_copy[tag]

    # profiles are automatically saved to disk when initialized
    assert user_copy._root_path.is_file()
    assert user_copy._profile_tokens._salt_path.is_file()
    assert (user_copy.path / user_copy.filename(tag)).is_file()

    # async & sync profile contructors are equivalent
    async_user = await AsyncDatabase.agenerate_profile(
        **PROFILE_AND_SETTINGS, preload=True
    )
    assert async_user[tag] == user[tag]
    assert async_user.tags == user.tags
    assert async_user[tag]
    assert async_user[tag] == test_data
    assert async_user[tag].__class__ is dict
    assert async_user[tag] is not user[tag]

    # deleting a profile removes its files from the filesystem
    user.delete_database()
    assert not user_copy._root_path.is_file()
    assert not user_copy._profile_tokens._salt_path.is_file()
    assert not (user_copy.path / user_copy.filename(tag)).is_file()


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A configuration type for `RawGUID`.
"""

__all__ = ["RawGUIDConfig", "RawGUIDContainer"]


import io
import warnings
from secrets import token_bytes

from aiootp._typing import Typing as t
from aiootp._constants import NANOSECONDS, BIG
from aiootp._exceptions import Issue
from aiootp.commons import Config, OpenFrozenSlots
from aiootp.asynchs import Clock


if not Clock(NANOSECONDS).has_adequate_resolution():
    warnings.warn(  # pragma: no cover
        f"\nBEWARE: Nanosecond clocks require better time resolution than "
        f"what's available from the system's time resolution of "
        f"{Clock._SYSTEM_TIME_RESOLUTION} seconds/tick. \nSeveral of the "
        f"package's tools, such as GUIDs, need fine resolution to function "
        f"correctly & satisfy their stated guarantees."
    )


class RawGUIDConfig(Config):
    """
    A configuration type for `RawGUID`.
    """

    __slots__ = (
        "TIMESTAMP_BYTES",
        "PRF_BYTES",
        "NODE_ID_BYTES",
        "TICKER_BYTES",
        "TICK_DOMAIN",
        "SIZE",
        "clock",
        "prf",
    )

    slots_types: t.Mapping[str, type] = dict(
        TIMESTAMP_BYTES=int,
        PRF_BYTES=int,
        NODE_ID_BYTES=int,
        TICKER_BYTES=int,
        TICK_DOMAIN=int,
        SIZE=int,
        clock=t.ClockType,
        prf=t.Callable,
    )

    def _process_size(self, size: int) -> int:
        """
        Ensures the specified raw guid size matches the combined length
        of its components.
        """
        if size != (
            self.TIMESTAMP_BYTES
            + self.PRF_BYTES
            + self.NODE_ID_BYTES
            + self.TICKER_BYTES
        ):
            raise Issue.invalid_length("stated guid size", size)
        return size

    def _process_clock(self, clock: t.Optional[type]) -> type:
        """
        Returns a default clock if one isn't provided.
        """
        if clock is None:
            return Clock(NANOSECONDS)
        else:
            return clock

    def __init__(
        self,
        *,
        timestamp_bytes: int,
        prf_bytes: int,
        node_id_bytes: int,
        ticker_bytes: int,
        size: int,
        clock: t.Optional[t.ClockType] = None,
        prf: t.Callable[[int], bytes] = token_bytes,
    ) -> None:
        self.TIMESTAMP_BYTES = timestamp_bytes
        self.PRF_BYTES = prf_bytes
        self.NODE_ID_BYTES = node_id_bytes
        self.TICKER_BYTES = ticker_bytes
        self.TICK_DOMAIN = int(self.TICKER_BYTES * "ff", 16)
        self.SIZE = self._process_size(size)
        self.clock = self._process_clock(clock)
        self.prf = prf


class RawGUIDContainer(OpenFrozenSlots):
    """
    Parses `RawGUID` bytes values into instance attributes.
    """

    __slots__ = ("timestamp", "token", "node_id", "ticker")

    _MAPPED_ATTRIBUTES: t.Tuple[str] = __slots__

    def __init__(self, guid: bytes, *, config: RawGUIDConfig) -> None:
        if not len(guid) == config.SIZE:
            raise Issue.invalid_length("raw guid", config.SIZE)
        reader = io.BytesIO(guid).read
        self.timestamp = reader(config.TIMESTAMP_BYTES)
        self.token = reader(config.PRF_BYTES)
        self.node_id = reader(config.NODE_ID_BYTES)
        self.ticker = reader(config.TICKER_BYTES)

    def __iter__(self) -> t.Generator[str, None, None]:
        yield from self._MAPPED_ATTRIBUTES

    def __hash__(self) -> int:
        return int.from_bytes(self.sort_key, BIG)

    def __eq__(self, other: t.Cls) -> bool:
        return self.sort_key == other.sort_key

    def __gt__(self, other: t.Cls) -> bool:
        return self.sort_key > other.sort_key

    def __lt__(self, other: t.Cls) -> bool:
        return self.sort_key < other.sort_key

    @property
    def sort_key(self) -> bytes:
        return self.timestamp + self.node_id + self.ticker + self.token


module_api = dict(
    RawGUIDConfig=t.add_type(RawGUIDConfig),
    RawGUIDContainer=t.add_type(RawGUIDContainer),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Improves the efficiency of the `AffineXORChain` by chaining
duplicates of the same permutation state.
"""

__all__ = ["FastAffineXORChain"]


from aiootp._typing import Typing as t
from aiootp.commons import ConfigMap

from .affine_permutation import AffinePermutation
from .affine_xor_chain_config import AffineXORChainConfig
from .affine_xor_chain import AffineXORChain


class FastAffineXORChain(AffineXORChain):
    """
    A more efficient subclass of the `AffineXORChain` permutation which
    saves on memory, time & cache miss costs by chaining together calls
    the same affine permutation instead of three independent ones.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp._permutations import FastAffineXORChain

    domain_size = 1
    key = token_bytes(FastAffineXORChain.key_size(domain_size))
    aff = FastAffineXORChain(key=key, config_id=domain_size)

    latin_square = []
    domain = tuple(range(256 ** domain_size))
    for _ in domain:
        row = tuple(aff.permute(element) for element in domain)
        latin_square.append(row)
        aff.step()

    assert len(set(latin_square)) == len(domain)
    assert all(len(set(row)) == len(domain) for row in latin_square)
    columns = [[row[column] for row in latin_square] for column in domain]
    assert all(len(set(column)) == len(domain) for column in columns)
    """

    __slots__ = ()

    _configs = ConfigMap(
        mapping={
            config_id: AffineXORChainConfig(
                config_id=config_id,
                size=config_id,
                permutation_type=AffinePermutation,
                permutation_config_id=config_id,
                key_types=AffineXORChainConfig._FAST_CHAIN_KEY_TYPES,
            )
            for config_id in [*range(1, 33), 64, 128, 192, 256]
        },
        config_type=AffineXORChainConfig,
    )

    def _initialize_permutations(
        self, key_reader: t.Callable[[int], bytes], size: int
    ) -> t.Tuple[t.PermutationType]:
        """
        Uses the permutation configuration ID saved in the instance's
        config to initialize & return three pointers to the same affine
        permutation object, which will be chained together. This
        overwrites the superclass' usage of this method in a compliant
        way while customizing the three pointers returned, enabling the
        improvement in efficiency.
        """
        cid = self.config.PERMUTATION_CONFIG_ID
        Permutation = self.config.Permutation
        aff = Permutation(key=key_reader(size), config_id=cid)
        return (aff, aff, aff)


module_api = dict(
    FastAffineXORChain=t.add_type(FastAffineXORChain),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Data processing & encoding utilities.
"""

__all__ = [
    "Domains",
    "abytes_are_equal",
    "ahash_bytes",
    "bytes_are_equal",
    "hash_bytes",
]


from .transform import *
from .canon import *
from .domains import *
from .byte_io import *
from .hashing import *


modules = dict(
    transform=transform,
    canon=canon,
    domains=domains,
    byte_io=byte_io,
    hashing=hashing,
)


module_api = dict(
    ByteIO=ByteIO,
    Domains=Domains,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    abytes_are_equal=abytes_are_equal,
    acanonical_pack=acanonical_pack,
    acanonical_unpack=acanonical_unpack,
    ahash_bytes=ahash_bytes,
    bytes_are_equal=bytes_are_equal,
    canonical_pack=canonical_pack,
    canonical_unpack=canonical_unpack,
    hash_bytes=hash_bytes,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *
from test_Clock import TIME_RESOLUTION

from aiootp.randoms.ids.raw_guid_config import RawGUIDContainer


class TestGUIDConfig:
    async def test_permutation_cid_can_be_specified(self) -> None:
        for permutation_config_id in (14, 15, 17, 20):
            config = t.GUIDConfig(
                config_id=16,
                size=16,
                raw_guid_type=t.RawGUID,
                permutation_type=t.FastAffineXORChain,
                permutation_config_id=permutation_config_id,
            )
            assert permutation_config_id == config.PERMUTATION_CONFIG_ID

    async def test_raw_cid_can_be_specified(self) -> None:
        for raw_guid_config_id in (14, 15, 17, 20):
            config = t.GUIDConfig(
                config_id=16,
                size=16,
                raw_guid_type=t.RawGUID,
                permutation_type=t.FastAffineXORChain,
                raw_guid_config_id=raw_guid_config_id,
            )
            assert raw_guid_config_id == config.RAW_GUID_CONFIG_ID


class TestGUID:
    async def test_unmasking_inverts_the_applied_permutation(self) -> None:
        for size in range(12, 33):
            key = token_bytes(GUID.key_size(size))
            guid = GUID(key=key, config_id=size)

            # ASYNC
            masked_guid = await guid.anew()
            unmasked_guid = await guid.aread(masked_guid)

            reconstructed_rawguid = b"".join(unmasked_guid.values())
            int_reconstructed_rawguid = int.from_bytes(
                reconstructed_rawguid, BIG
            )
            permuted_rawguid = await guid._permutation.apermute(
                int_reconstructed_rawguid
            )
            assert masked_guid == permuted_rawguid.to_bytes(size, BIG)

            # SYNC
            masked_guid = guid.new()
            unmasked_guid = guid.read(masked_guid)

            reconstructed_rawguid = b"".join(unmasked_guid.values())
            int_reconstructed_rawguid = int.from_bytes(
                reconstructed_rawguid, BIG
            )
            permuted_rawguid = guid._permutation.permute(
                int_reconstructed_rawguid
            )
            assert masked_guid == permuted_rawguid.to_bytes(size, BIG)

    async def test_raw_guids_sorting(self) -> None:
        runs = 32
        indexes = list(range(runs))
        for size in range(12, 33):
            key = token_bytes(GUID.key_size(size))
            guid = GUID(key=key, config_id=size)
            raw_guids = {guid.read(guid.new()) for _ in indexes}
            assert runs == len(raw_guids)
            raw_guids_list = sorted(raw_guids)
            for i, raw in enumerate(raw_guids_list):
                if i < runs - 1:
                    assert raw != raw_guids_list[i + 1]
                    assert raw < raw_guids_list[i + 1]
                else:
                    assert raw != raw_guids_list[i - 1]
                    assert raw > raw_guids_list[i - 1]

    async def test_guid_size_limits(self) -> None:
        problem = (  # fmt: skip
            "Invalid length GUID was allowed during decoding."
        )
        for size in range(12, 33):
            key = token_bytes(GUID.key_size(size))
            guid = GUID(key=key, config_id=size)
            with Ignore(ValueError, if_else=violation(problem)):
                guid.read(b"\xff" + guid.new())
            with Ignore(ValueError, if_else=violation(problem)):
                RawGUIDContainer(b"\xff" + guid.new(), config=guid.config)
            with Ignore(ValueError, if_else=violation(problem)):
                await guid.aread(b"\xff" + guid.new())

    async def test_node_id_size_limits(self) -> None:
        problem = (  # fmt: skip
            "An invalid length node ID was allowed."
        )
        for size in range(12, 33):
            key = token_bytes(GUID.key_size(size))
            config = GUID._configs[size]
            rg_config = config.RawGUID._configs[config.RAW_GUID_CONFIG_ID]
            with Ignore(ValueError, if_else=violation(problem)):
                GUID(
                    key=key,
                    config_id=size,
                    node_id=(rg_config.NODE_ID_BYTES + 1) * "ff",
                )

    async def test_guid_uniqueness(self) -> None:
        size = 12
        key = token_bytes(GUID.key_size(size))
        guid = GUID(key=key, config_id=size)

        runs = 2048
        # PROBLEM: Low resolution system clock harms GUID uniqueness guarantees.
        is_os_clock_resolution_issue = lambda _: (TIME_RESOLUTION >= 1e-04)

        # ASYNC
        guids = {await guid.anew() for _ in range(runs)}
        with Ignore(AssertionError, if_except=is_os_clock_resolution_issue):
            assert runs == len(guids)

        # SYNC
        for _ in range(runs):
            guids.add(guid.new())
        with Ignore(AssertionError, if_except=is_os_clock_resolution_issue):
            assert 2 * runs == len(guids)


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})
